#!/usr/bin/python3

# Copyright (C) 2020  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

from dateutil import parser
from typing import List
from lofar.sas.tmss.tmss.tmssapp.models import *
from lofar.sas.tmss.tmss.tmssapp.conversions import antennafields_for_antennaset_and_station
from lofar.parameterset import parameterset
from lofar.common.util import single_line_with_single_spaces
from lofar.sas.tmss.tmss.exceptions import SubtaskInvalidStateException, SubtaskException

import logging
logger = logging.getLogger(__name__)

def parse_parset_vector(parset_value: str, parameterset_func=parameterset.getIntVector) -> List[int]:
    """ Parse a parameterset vector, by passing it through the parameterset class.

        Provide the actual function the parameterset class uses to get the vector. """
    p = parameterset()
    p.add("key", parset_value)
    return parameterset_func(p, "key", expandable=True)

def parse_antennafield_name(name: str) -> dict:
    """ Parse a name like CS001HBA into a dict denoting the station, field, and type as used in the feedback_doc. """

    if name.endswith("HBA"):
        return { "station": name[:-3], "field": "HBA", "type": "HBA" }
    elif name.endswith("LBA"):
        return { "station": name[:-3], "field": "LBA", "type": "LBA" }
    elif name.endswith("HBA0"):
        return { "station": name[:-4], "field": "HBA0", "type": "HBA" }
    elif name.endswith("HBA1"):
        return { "station": name[:-4], "field": "HBA1", "type": "HBA" }
    else:
        raise Exception("Invalid station name supplied: %s" % name)


def translate_storagewriter_parset_to_doc(storagewriter: str) -> str:
    """ Translate a storagewriter as mentioned in the feedback parset to one we can use in the feedback_doc. """
    storagewriter = storagewriter.lower()

    if storagewriter == "casa":
        return "standard"
    elif storagewriter == "lofar":
        return "lofarstman"
    elif storagewriter == "dysco":
        return "dysco"
    elif storagewriter == "hdf5default":
        return "standard"
    else:
        return storagewriter

def observation_correlated_feedback_to_feedback_doc(dp_feedback: dict) -> dict:
    """ Transform feedback for Correlated dataproducts from observations into a feedback_doc dictionary."""

    # This feedback is generated in RTCP/Cobalt/CoInterface/src/LTAFeedback.cc

    feedback_doc = {
        "percentage_written": int(dp_feedback['percentageWritten']),
        "frequency": {
            "subbands": [int(dp_feedback['stationSubband'])],
            "central_frequencies": [float(dp_feedback['centralFrequency'])],
            "channel_width": float(dp_feedback['channelWidth']),
            "channels_per_subband": int(dp_feedback['channelsPerSubband']),
        },
        "time": {
            "start_time": parser.parse(dp_feedback['startTime'], ignoretz=True).isoformat()+'Z',
            "duration": float(dp_feedback['duration']),
            "sample_width": float(dp_feedback['integrationInterval']),
        },
        "antennas": {
            "set": dp_feedback['antennaSet'],
            "fields": [parse_antennafield_name(field) for field in parse_parset_vector(dp_feedback['antennaFields'], parameterset.getStringVector)],
        },
        "target": {
            "pointing": {
                "direction_type": dp_feedback['Pointing.directionType'],
                "angle1": float(dp_feedback['Pointing.angle1']),
                "angle2": float(dp_feedback['Pointing.angle2']),
                "target": "Unknown"  # todo
            }
        },
        "samples": {
            "polarisations": ["XX","XY","YX","YY"],
            "type": "float",
            "bits": 32,
            "writer": translate_storagewriter_parset_to_doc(dp_feedback['storageWriter']),
            "writer_version": dp_feedback['storageWriterVersion'],
            "complex": True
        }
    }

    return feedback_doc

def observation_beamformed_feedback_to_feedback_doc(dp_feedback: dict) -> dict:
    """ Transform feedback for Beamformed dataproducts from observations into a feedback_doc dictionary."""

    # This feedback is generated in RTCP/Cobalt/CoInterface/src/LTAFeedback.cc

    # Determine which type of beam we're dealing with, to address the right array.
    # TODO: Remove this array layer from the feedback in COBALT
    beam_prefix = dp_feedback["beamPrefix"] # f.e. CoherentStokesBeam[0].

    feedback_doc = {
        "percentage_written": int(dp_feedback['percentageWritten']),
        "frequency": {
            "subbands": parse_parset_vector(dp_feedback[beam_prefix + 'stationSubbands'], parameterset.getIntVector),
            "central_frequencies": parse_parset_vector(dp_feedback[beam_prefix + 'centralFrequencies'], parameterset.getDoubleVector),
            "channel_width": float(dp_feedback[beam_prefix + 'channelWidth']),
            "channels_per_subband": int(dp_feedback[beam_prefix + 'channelsPerSubband']),
        },
        "time": {
            "start_time": parser.parse(dp_feedback['startTime'], ignoretz=True).isoformat()+'Z',
            "duration": float(dp_feedback['duration']),
            "sample_width": float(dp_feedback[beam_prefix + 'samplingTime']),
        },
        "antennas": {
            "set": dp_feedback[beam_prefix + 'antennaSet'],
            "fields": [parse_antennafield_name(field) for field in parse_parset_vector(dp_feedback[beam_prefix + 'antennaFields'], parameterset.getStringVector)],
        },
        "target": {
            "pointing": {
                "direction_type": dp_feedback[beam_prefix + 'Pointing.directionType'],
                "angle1": float(dp_feedback[beam_prefix + 'Pointing.angle1']),
                "angle2": float(dp_feedback[beam_prefix + 'Pointing.angle2']),
                "target": "Unknown"  # todo
            },
            "coherent": not beam_prefix.startswith("IncoherentStokesBeam")
        },
        "samples": {
            "polarisations": parse_parset_vector(dp_feedback[beam_prefix + 'stokes'], parameterset.getStringVector),
            "type": dp_feedback[beam_prefix + 'sampleType'],
            "bits": int(dp_feedback[beam_prefix + 'sampleNrBits']),
            "writer": translate_storagewriter_parset_to_doc(dp_feedback['storageWriter']),
            "writer_version": dp_feedback['storageWriterVersion'],
            "complex": False # Complex Voltages write complex values, but split real and imaginary values into separate files
        }
    }

    return feedback_doc

def preprocessing_pipeline_feedback_to_feedback_doc(input_dp_feedback_doc: dict, dp_feedback: dict) -> dict:
    """ Transform feedback for Correlated dataproducts from preprocessing pipelines into a feedback_doc dictionary."""

    # This feedback is generated in CEP/Pipeline/recipes/sip/helpers/metadata.py

    feedback_doc = {
        "percentage_written": int(dp_feedback['percentageWritten']),
        "frequency": {
            "subbands": input_dp_feedback_doc['frequency']['subbands'], # pipeline feedback always reports 0 as the stationSubband, unfortunately
            "central_frequencies": [float(dp_feedback['centralFrequency'])],
            "channel_width": float(dp_feedback['channelWidth']),
            "channels_per_subband": int(dp_feedback['channelsPerSubband']),
        },
        "time": {
            "start_time": parser.parse(dp_feedback['startTime'], ignoretz=True).isoformat()+'Z',
            "duration": float(dp_feedback['duration']),
            "sample_width": float(dp_feedback['integrationInterval']),
        },
        "antennas": {
            "set": input_dp_feedback_doc["antennas"]['set'],
            "fields": input_dp_feedback_doc["antennas"]['fields'],
        },
        "target": {
            "pointing": input_dp_feedback_doc["target"]["pointing"],
        },
        "samples": {
            "polarisations": ["XX","XY","YX","YY"],
            "type": "float",
            "bits": 32,
            "writer": translate_storagewriter_parset_to_doc(dp_feedback['storageWriter']),
            "writer_version": dp_feedback['storageWriterVersion'],
            "complex": True
        }
    }

    return feedback_doc


def pulsar_pipeline_summary_feedback_to_feedback_doc(dp_feedback: dict) -> dict:
    """ Transform feedback for pulsar pipeline "summary" dataproducts from observations into a feedback_doc dictionary."""

    # This feedback is generated in https://git.astron.nl/ro/pulp/-/blob/master/pulp_feedback.py

    feedback_doc = {
        "percentage_written": int(dp_feedback['percentageWritten']),
        # set of files must be unique, but PULP sends us duplicates 
        "files": list(set(parse_parset_vector(dp_feedback['fileContent'], parameterset.getStringVector))),
        "target": {
            "coherent": dp_feedback['datatype'] != "SummaryIncoherentStokes"
        }
    }

    return feedback_doc


def pulsar_pipeline_analysis_feedback_to_feedback_doc(input_dp_feedback_doc: dict, dp_feedback: dict) -> dict:
    """ Transform feedback for pulsar pipeline "analysis" dataproducts from observations into a feedback_doc dictionary.
        
        input_dp_feedback_doc can be any input dataproduct related to this output dataproduct. Typically,
        there is only 1, but in case of ComplexVoltages, there are 4 inputs (each of Xre/Xim/Yre/Yim). """

    # This feedback is generated in https://git.astron.nl/ro/pulp/-/blob/master/pulp_feedback.py

    for beam_prefix in ["FlysEyeBeam.", "CoherentStokesBeam.", "IncoherentStokesBeam."]:
        if any(k.startswith(beam_prefix) for k in dp_feedback.keys()):
            break
    else:
        raise ValueError("Could not determine beam prefix. Feedback: %s" % dp_feedback)

    feedback_doc = {
        "percentage_written": int(dp_feedback['percentageWritten']),
        # set of files must be unique, but PULP sends us duplicates 
        "files": list(set(parse_parset_vector(dp_feedback['fileContent'], parameterset.getStringVector))),
        "frequency": {
            "subbands": parse_parset_vector(dp_feedback[beam_prefix + 'stationSubbands'], parameterset.getIntVector),
            "central_frequencies": parse_parset_vector(dp_feedback[beam_prefix + 'centralFrequencies'], parameterset.getDoubleVector),
            "channel_width": float(dp_feedback[beam_prefix + 'channelWidth']),
            "channels_per_subband": int(dp_feedback[beam_prefix + 'channelsPerSubband']),
        },
        "time": {
            "start_time": input_dp_feedback_doc["time"]["start_time"],
            "duration":  input_dp_feedback_doc["time"]["duration"],
            "sample_width": float(dp_feedback[beam_prefix + 'samplingTime']),
        },
        "antennas": {
            "set": input_dp_feedback_doc["antennas"]['set'],
            "fields": [parse_antennafield_name(dp_feedback[beam_prefix + 'stationName'] + dp_feedback[beam_prefix + 'antennaFieldName'])] if beam_prefix == "FlysEyeBeam." else input_dp_feedback_doc["antennas"]['fields'],
        },
        "target": {
            "pointing": {
                "direction_type": "J2000",
                "angle1": float(dp_feedback[beam_prefix + 'Pointing.angle1']),
                "angle2": float(dp_feedback[beam_prefix + 'Pointing.angle2']),
                "target": input_dp_feedback_doc["target"]["pointing"]["target"]
            } if beam_prefix == "CoherentStokesBeam." else input_dp_feedback_doc["target"]["pointing"],
            "coherent": beam_prefix != "IncoherentStokesBeam."
        },
        "samples": {
            "polarisations": parse_parset_vector(dp_feedback[beam_prefix + 'stokes'], parameterset.getStringVector),
            "type": input_dp_feedback_doc["samples"]["type"],
            "bits": input_dp_feedback_doc["samples"]["bits"],
            "writer": "standard",
            "writer_version": "UNKNOWN",
            "complex": dp_feedback['datatype'] == "ComplexVoltages",
        }
    }

    return feedback_doc


def extract_dataproduct_feedback(feedback: parameterset) -> dict:
    """ Returns a dict of the feedback strings per dataproduct, extracted from a feedback parset. Example:

        extract_dataproduct_feedback(
            parameterset.fromString("LOFAR.ObsSW.Observation.Dataproducts.Output_Beamformed_[0].key1=value1")
        ) ==
        { "Output_Beamformed_[0]": { "key1": "value1" } }
    """

    # locate where the dataproducts are, to cut off any prefix like "LOFAR.ObsSW."
    dp_feedback = feedback.makeSubset(feedback.fullModuleName('Observation.DataProducts') + '.')

    # extract the unique dataproduct keys (f.e. "Output_Beamformed_[0]"), so we can loop over them
    dp_keys = sorted(list(set([key[:key.find('.')] for key in dp_feedback.keys() if key.startswith('Output_')])))

    # extract all info for each dataproduct
    return {dp_key: dp_feedback.makeSubset(dp_key+'.').dict() for dp_key in dp_keys}


def process_feedback_into_subtask_dataproducts(subtask:Subtask, feedback: parameterset) -> Subtask:
    if subtask.specifications_template.type.value not in [SubtaskType.Choices.OBSERVATION.value, SubtaskType.Choices.PIPELINE.value]:
            raise SubtaskException("Cannot process feedback for subtask id=%s since type=%s not in %s" %
                                   (subtask.id, subtask.specifications_template.type.value,
                                    [SubtaskType.Choices.OBSERVATION.value, SubtaskType.Choices.PIPELINE.value]))

    if subtask.state.value != SubtaskState.objects.get(value='finishing').value:
        raise SubtaskInvalidStateException("Cannot process feedback for subtask id=%s because the state is '%s' and not '%s'" % (subtask.id, subtask.state.value, SubtaskState.Choices.FINISHING.value))

    logger.info('processing feedback into the dataproducts of subtask id=%s type=%s feedback:\n%s', subtask.id, subtask.specifications_template.type.value, str(feedback))

    # process each dataproduct subset
    for dp_key, dp_feedback in extract_dataproduct_feedback(feedback).items():
        # determine corresponding TMSS dataproduct
        try:
            dataproduct = subtask.output_dataproducts.get(filename=dp_feedback.get('filename'))
        except Exception as e:
            # process reported about a file we didn't know it would write!

            logger.exception("cannot process feedback: %s. No output dataproduct known for subtask id=%s feedback: %s", e, subtask.id, dp_feedback)
            continue

        try:
            logger.info('processing feedback for dataproduct id=%s filename=%s of subtask id=%s feedback: %s', dataproduct.id, dataproduct.filename, subtask.id, single_line_with_single_spaces(str(dp_feedback)))

            # derive values or collect for different subtask types
            if subtask.specifications_template.type.value == SubtaskType.Choices.OBSERVATION.value:
                if dataproduct.datatype.value == "visibilities":
                    dataproduct.feedback_template = DataproductFeedbackTemplate.objects.get(name='feedback')
                    dataproduct.feedback_doc = observation_correlated_feedback_to_feedback_doc(dp_feedback)
                elif dataproduct.datatype.value == "time series":
                    dataproduct.feedback_template = DataproductFeedbackTemplate.objects.get(name='feedback')
                    dataproduct.feedback_doc = observation_beamformed_feedback_to_feedback_doc(dp_feedback)
                else:
                    logger.error("cannot process feedback for dataproduct with specification template %s", dataproduct.specifications_template.name)

            elif subtask.specifications_template.type.value == SubtaskType.Choices.PIPELINE.value:
                if subtask.specifications_template.name == "preprocessing pipeline":
                    # each file has exactly 1 input.
                    input_dataproduct = DataproductTransform.objects.get(output=dataproduct).input

                    dataproduct.feedback_template = DataproductFeedbackTemplate.objects.get(name='feedback')
                    dataproduct.feedback_doc = preprocessing_pipeline_feedback_to_feedback_doc(input_dataproduct.feedback_doc, dp_feedback)

                elif subtask.specifications_template.name == "pulsar pipeline":
                    if dataproduct.specifications_template.name == "pulp summary":
                        # pulp summary file
                        dataproduct.feedback_template = DataproductFeedbackTemplate.objects.get(name='pulp summary')
                        dataproduct.feedback_doc = pulsar_pipeline_summary_feedback_to_feedback_doc(dp_feedback)
                    else:
                        # pulp analysis file, which has 1-4 inputs. just use the first.
                        input_dataproduct = DataproductTransform.objects.filter(output=dataproduct).first().input

                        dataproduct.feedback_template = DataproductFeedbackTemplate.objects.get(name='feedback')
                        dataproduct.feedback_doc = pulsar_pipeline_analysis_feedback_to_feedback_doc(input_dataproduct.feedback_doc, dp_feedback)
                else:
                    logger.error("cannot process feedback for pipelines with schema named %s", subtask.specifications_template.name)

            # set value of dataproduct size field
            dataproduct.size = dp_feedback.get('size', dataproduct.size)

            # commit to database
            dataproduct.save()
            logger.info('saved processed feedback into dataproduct id=%s filename=%s feedback_doc=%s', dataproduct.id, dataproduct.filename, dataproduct.feedback_doc)
        except Exception as e:
            logger.exception('error while processing feedback for dataproduct id=%s filename=%s feedback=%s', dataproduct.id, dataproduct.filename, dp_feedback)

    subtask.refresh_from_db()
    return subtask


def append_to_subtask_raw_feedback(subtask: Subtask, feedback: parameterset) -> Subtask:
    """ append/merge the given feedback into the already stored raw_feedback
    by using this parset-merging strategy we ensure a properly formatted plain text raw_feedback doc with no duplicate entries
    """
    # check we are in finishing state
    if subtask.state.value != SubtaskState.objects.get(value='finishing').value:
        raise SubtaskInvalidStateException("Cannot process feedback for subtask id=%s because the state is '%s' and not '%s'" % (subtask.id, subtask.state.value, SubtaskState.Choices.FINISHING.value))

    feedback_parset = parameterset.fromString(subtask.raw_feedback or "")
    feedback_parset.adoptDict(feedback.dict())
    subtask.raw_feedback = str(feedback_parset)
    subtask.save()
    return subtask

def process_feedback_for_subtask_and_set_to_finished_if_complete(subtask: Subtask, feedback_doc: str) -> Subtask:
    """
    Translates raw feedback from a subtask (which has been provided by Cobalt or pipelines) and translate it to
    json documents for the individual dataproducts.
    """
    if subtask.state.value == SubtaskState.objects.get(value='started').value:
        logger.info("received feedback for subtask id=%s while it is still running (state=%s). Setting state to 'finishing' and stop_time to now.", subtask.id, subtask.state.value)
        subtask.state = SubtaskState.objects.get(value='finishing')
        subtask.scheduled_on_sky_stop_time = datetime.utcnow()
        subtask.save()

    # the submitted feedback_doc is (should be) a plain text document in parset format
    # so, treat it as a parset
    new_feedback_parset = parameterset.fromString(feedback_doc)

    # store the new feedback
    subtask = append_to_subtask_raw_feedback(subtask=subtask, feedback=new_feedback_parset)

    # and process it
    subtask = process_feedback_into_subtask_dataproducts(subtask, new_feedback_parset)

    # if complete, set subtask state to finished
    if subtask.is_feedback_complete:
        logger.info("Feedback for subtask id=%s is complete. Setting state to 'finished'", subtask.id)
        subtask.state = SubtaskState.objects.get(value='finished')
        subtask.save()

    return subtask


def reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete(subtask: Subtask) -> Subtask:
    """
    Reprocesses the stored raw feedback from a subtask (which has been provided by Cobalt or pipelines) and translate it to
    json documents for the individual dataproducts.
    """
    if not subtask.raw_feedback:
        raise SubtaskException("Cannot reprocess raw_feedback for subtask id=%s because it is empty" % (subtask.id))

    try:
        raw_feedback_parset = parameterset.fromString(subtask.raw_feedback)
    except Exception as e:
        raise SubtaskException("Cannot reprocess raw_feedback for subtask id=%s because it cannot be interpreted as parset. Error=%s\n%s" % (subtask.id, str(e), subtask.raw_feedback))

    if subtask.state.value in (SubtaskState.objects.get(value='started').value, SubtaskState.objects.get(value='finished').value):
        # allow started/running subtask to go to finishing, (re)process, and then go to finished if complete
        # allow finished subtask to go back to finishing, reprocess, and then go to finished again if complete
        subtask.state = SubtaskState.objects.get(value='finishing')
        subtask.save()

    subtask = process_feedback_into_subtask_dataproducts(subtask, raw_feedback_parset)

    # if complete, set subtask state to finished
    if subtask.is_feedback_complete:
        logger.info("Feedback for subtask id=%s is complete. Setting state to 'finished'", subtask.id)
        subtask.state = SubtaskState.objects.get(value='finished')

    subtask.save()
    return subtask


