from django.db.models import F, Sum
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp import serializers
from lofar.sas.tmss.tmss.exceptions import TaskOverlappingWarning
from lofar.sas.tmss.services.scheduling.constraints.template_constraints_v1 import get_target_observation_task_name_from_specifications_doc

from rest_framework.request import Request
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
import warnings

# Cycle Report


def create_cycle_report(request: Request, cycle: models.Cycle, start: datetime, stop: datetime) -> {}:
    """
    Create a cycle report as a JSON object.
    """
    with warnings.catch_warnings(record=True) as w:  # Catch potential TaskOverlappingWarning
        warnings.simplefilter('once', TaskOverlappingWarning)

        result = {'cycle': cycle.pk,
                  'telescope_time_distribution': _get_telescope_time_distribution(cycle, start, stop),
                  'average_efficiency': _get_average_efficiency(cycle, start, stop),
                  'completion_level': _get_completion_level(cycle, start, stop),
                  'observation_hours_per_category': _get_observation_hours_per_category(cycle, start, stop),
                  'weekly_efficiency': _get_weekly_efficiency(cycle, start, stop),
                  'data_ingested_per_site_and_category': _get_data_ingested_per_site_and_category(cycle, start, stop),
                  'projects_summary': _get_projects_summary(request, cycle, start, stop),
                  'usage_mode': _get_usage_mode(cycle, start, stop), 'failures': _get_failures(cycle, start, stop),}
        result['contains_overlapping_observations'] = any([issubclass(warning.category, TaskOverlappingWarning) for warning in w])

    return result


def _get_telescope_time_distribution(cycle: models.Cycle, start: datetime, stop: datetime) -> {}:
    """
    Help function to retrieve telescope time distribution info.
    """
    from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess

    result = {}

    # Consider UNASSIGNED and FILLER as categories for the purposes of reporting
    categories = ['UNASSIGNED', 'FILLER', ] + [c for c in models.ProjectCategory.Choices]
    for c in categories:
        total, succeeded, failed = 0, 0, 0
        projects = models.Project.objects.filter(cycles=cycle, project_category=c.value) if (c != 'UNASSIGNED' and c != 'FILLER') \
            else models.Project.objects.filter(cycles=cycle, filler=True) if c == 'FILLER' \
            else models.Project.objects.filter(cycles=cycle, project_category__isnull=True)
        for p in projects:
            # Get durations for single project
            subs = models.SchedulingUnitBlueprint.objects.filter(draft__scheduling_set__project=p.pk)
            for sub in subs:
                if sub.observed_duration and sub.observed_start_time >= start and sub.observed_end_time <= stop:
                    sup = SchedulingUnitProcess.objects.filter(su=sub).first()
                    # Aggregate total, successful and failed durations
                    total += sub.observed_duration.total_seconds()
                    succeeded += sub.observed_duration.total_seconds() if sup and sup.results_accepted else 0
                    failed += sub.observed_duration.total_seconds() if sup and sup.results_accepted is False else 0
        # Calculate derived and store durations
        idle = total - succeeded - failed
        succeeded_perc = round(succeeded / total, 2) if total > 0 else None
        failed_perc = round(failed / total, 2) if total > 0 else None
        idle_perc = round(idle / total, 2) if total > 0 else None
        durations = {'total': total, 'succeeded': succeeded, 'succeeded_perc': succeeded_perc, 'failed': failed, 'failed_perc': failed_perc, 'idle': idle, 'idle_perc': idle_perc}
        result[c if c == 'UNASSIGNED' or c == 'FILLER' else c.name] = {'durations': durations}

    return result


def _get_average_efficiency(cycle: models.Cycle, start: datetime, stop: datetime) -> {}:
    """
    Help function to retrieve the average efficiency with total and total successful obs durations per day.
    """
    from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess

    result = {'target': 0.65}  # TODO: Default efficiency is 65%. Change it properly when it will be implemented.
    efficiency_per_day = 0

    # Get SchedulingUnitBlueprints related to the cycle
    subs = models.SchedulingUnitBlueprint.objects.filter(draft__scheduling_set__project__cycles=cycle.pk)

    # Iterate through days and sum durations per day
    i, step, d = 0, timedelta(hours=24), start
    while d < stop:
        total_per_day = 0
        total_succeeded_per_day = 0
        for sub in subs:
            # Aggregate total and successful durations
            if sub.observed_duration and d <= sub.observed_start_time < d + step and sub.observed_end_time < d + step:
                sup = SchedulingUnitProcess.objects.filter(su=sub).first()
                total_per_day += sub.observed_duration.total_seconds()
                total_succeeded_per_day += sub.observed_duration.total_seconds() if sup and sup.results_accepted else 0
        efficiency_per_day += total_succeeded_per_day / total_per_day if total_per_day > 0 else 0
        i += 1
        d += step

    # Store the efficiency averaged over the entire cycle
    result['efficiency'] = round(efficiency_per_day / i, 2) if efficiency_per_day > 0 else None
    return result


def _get_completion_level(cycle: models.Cycle, start: datetime, stop: datetime) -> {}:
    """
    Help function to retrieve the completion level info.
    """
    from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess

    # TODO: Change it properly when it will be implemented.
    result = {'target': 0.0}

    # Get SchedulingUnitBlueprints related to the cycle
    subs = models.SchedulingUnitBlueprint.objects.filter(draft__scheduling_set__project__cycles=cycle.pk)

    total, total_succeeded = 0, 0
    for sub in subs:
        # Aggregate total and successful durations
        if sub.observed_duration and sub.observed_start_time >= start and sub.observed_end_time <= stop:
            sup = SchedulingUnitProcess.objects.filter(su=sub).first()
            total += sub.observed_duration.total_seconds()
            total_succeeded += sub.observed_duration.total_seconds() if sup and sup.results_accepted else 0
    # Store durations
    succeeded_perc = round(total_succeeded / total, 2) if total > 0 else None
    result['total'], result['succeeded'], result['succeeded_perc'] = total, total_succeeded, succeeded_perc

    # Calculate prognosis
    unschedulable_subtasks = models.Subtask.objects.filter(task_blueprint__scheduling_unit_blueprint__in=subs).filter(state='unschedulable')
    unschedulable_duration = sum([uns.duration.total_seconds() for uns in unschedulable_subtasks])
    result['prognosis'] = round(unschedulable_duration / total, 2) if total > 0 else None

    return result


def _get_observation_hours_per_category(cycle: models.Cycle, start: datetime, stop: datetime) -> {}:
    """
    Help function to retrieve observation hours per category info.
    """
    from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess

    result = {'total_duration': 0, 'total_duration_successful': 0, 'total_duration_failed': 0, 'total_duration_idle': 0, 'DDT Com Rep': 0}

    # Filter System Unavailability
    # TODO: Maybe add some other case scenarios where the system is not up and running?
    reservations_duration = models.Reservation.objects.filter(start_time__gte=start, stop_time__lte=stop).\
        filter(specifications_doc__activity__type='maintenance').aggregate(duration=Sum(F('stop_time') - F('start_time')))['duration']
    result['System Unavailability'] = reservations_duration.total_seconds() if reservations_duration else None

    # Filter durations for each prio basing on SUBs and projects
    cycle_subs = models.SchedulingUnitBlueprint.objects.filter(draft__scheduling_set__project__cycles=cycle.pk)
    for prio in models.PriorityQueueType.Choices:
        result[f'total_duration_successful_{prio.name}'] = 0
        subs = cycle_subs.filter(priority_queue=prio.value)
        for sub in subs:
            if sub.observed_duration and sub.observed_start_time >= start and sub.observed_end_time <= stop:
                # Filter DDT Com
                project_category = sub.project.project_category
                if project_category and (project_category.value == 'ddt' or project_category.value == 'commissioning'):
                    result['DDT Com Rep'] += sub.observed_duration.total_seconds()
                # Aggregate total and idle
                result['total_duration'] += sub.observed_duration.total_seconds()
                result['total_duration_idle'] += sub.observed_duration.total_seconds()
                # Distinguish successful and failed
                sup = SchedulingUnitProcess.objects.filter(su=sub).first()
                result[f'total_duration_successful_{prio.name}'] += sub.observed_duration.total_seconds() if sup and sup.results_accepted else 0
                result['total_duration_failed'] += sub.observed_duration.total_seconds() if sup and sup.results_accepted is False else 0
        # Calculate prio percentages
        result[f'successful_{prio.name}_perc'] = round(result[f'total_duration_successful_{prio.name}'] / result['total_duration'], 2) if result['total_duration'] > 0 else None
        # Aggregate prio durations to general successful
        result['total_duration_successful'] += result[f'total_duration_successful_{prio.name}']
        # Subtract prio states from total to get partial idle
        result['total_duration_idle'] -= result[f'total_duration_successful_{prio.name}']
    # Subtract total failed to get total idle eventually
    result['total_duration_idle'] -= result['total_duration_failed']

    # Calculate percentages
    result['successful_perc'] = round(result['total_duration_successful'] / result['total_duration'], 2) if result['total_duration'] > 0 else None
    result['idle_perc'] = round(result['total_duration_idle'] / result['total_duration'], 2) if result['total_duration'] > 0 else None
    result['failed_perc'] = round(result['total_duration_failed'] / result['total_duration'], 2) if result['total_duration'] > 0 else None
    result['ddt_com_rep_perc'] = round(result['DDT Com Rep'] / result['total_duration'], 2) if result['total_duration'] > 0 else None
    result['system_unavailability_perc'] = round(result['System Unavailability'] / result['total_duration'], 2) if result['System Unavailability'] and result['total_duration'] > 0 else None

    return result


def _get_weekly_efficiency(cycle: models.Cycle, start: datetime, stop: datetime) -> {}:
    """
    Help function to retrieve the weekly efficiency with total successful obs durations per week.
    """
    from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess

    result = {'weeks': []}

    # Get SchedulingUnitBlueprints related to the cycle
    subs = models.SchedulingUnitBlueprint.objects.filter(draft__scheduling_set__project__cycles=cycle.pk)

    # Iterate through weeks and sum durations per week
    step, d = timedelta(days=7), start
    while d < stop:
        total_per_week = 0
        total_succeeded_per_week = 0
        for sub in subs:
            # Aggregate total and successful durations
            if sub.observed_duration and d <= sub.observed_start_time < d + step and sub.observed_end_time < d + step:
                sup = SchedulingUnitProcess.objects.filter(su=sub).first()
                total_per_week += sub.observed_duration.total_seconds()
                total_succeeded_per_week += sub.observed_duration.total_seconds() if sup and sup.results_accepted else 0
        efficiency = round(total_succeeded_per_week / total_per_week, 2) if total_per_week > 0 else None
        result['weeks'].append({'week': d.date().isoformat(), 'efficiency': efficiency})
        d += step

    return result


def _get_data_ingested_per_site_and_category(cycle: models.Cycle, start: datetime, stop: datetime) -> {}:
    """
    Help function to retrieve data ingested per site and category info.
    """
    result = {}

    # TODO: Currently there is no way to fitler per LTA site.
    # # Get DataProducts related to the cycle with an ArchiveInfo
    # archive_info = models.DataproductArchiveInfo.objects.filter(dataproduct__producer__subtask__task_blueprint__draft__scheduling_unit_draft__scheduling_set__project__cycles=cycle.pk)
    # dataproducts = [ai.dataproduct for ai in archive_info]

    # Get DataProducts related to the cycle from Subtasks with state 'finished'
    dataproducts = models.Dataproduct.objects.filter(producer__subtask__state='finished', producer__subtask__task_blueprint__draft__scheduling_unit_draft__scheduling_set__project__cycles=cycle.pk, producer__subtask__scheduled_on_sky_start_time__gte=start, producer__subtask__scheduled_on_sky_stop_time__lte=stop)

    # Combine and filter DataProducts accordingly
    dps_interferometric_obs_sizes = dataproducts.filter(producer__subtask__specifications_template__type='observation', dataformat='MeasurementSet', datatype='visibilities').aggregate(Sum('size'))
    result['Interferometric Observation'] = dps_interferometric_obs_sizes

    dps_beamformed_obs_sizes = dataproducts.filter(producer__subtask__specifications_template__type='observation', dataformat='Beamformed', datatype='time series').aggregate(Sum('size'))
    result['Beamformed Observation'] = dps_beamformed_obs_sizes

    dp_averaging_pip_sizes = dataproducts.filter(producer__subtask__specifications_template__type='pipeline', dataformat='MeasurementSet', datatype='visibilities').aggregate(Sum('size'))
    result['Preprocessing Pipeline'] = dp_averaging_pip_sizes

    dp_pulsar_pip_sizes = dataproducts.filter(producer__subtask__specifications_template__type='pipeline', dataformat__in=['pulp summary', 'pulp analysis'], datatype__in=['time series', 'pulsar profile']).aggregate(Sum('size'))
    result['Pulsar Pipeline'] = dp_pulsar_pip_sizes

    return result


def _get_projects_summary(request: Request, cycle: models.Cycle, start: datetime, stop: datetime) -> {}:
    """
    Help function to retrieve projects info.
    """
    projects_summary = []
    # Get Projects related to the Cycle
    projects = models.Project.objects.filter(cycles=cycle)
    for p in projects:
        # Include the Project report info
        projects_summary.append(create_project_report(request, p, start, stop))

    return projects_summary


def _get_usage_mode(cycle: models.Cycle, start: datetime, stop: datetime) -> {}:
    """
    Help function to retrieve ILT and local usage mode info.
    """
    result = {'all modes': {}, 'stand-alone mode': {}, 'ILT mode': {}}

    # Get all the reservations in the cycle semester
    reservations = models.Reservation.objects.filter(start_time__gte=start, stop_time__lte=stop)
    duration = reservations.aggregate(duration=Sum(F('stop_time') - F('start_time')))['duration']
    result['all modes']['total'] = duration.total_seconds() if duration else None
    # Production projects
    reservations_projects = reservations.filter(project__cycles=cycle.pk, project__project_category__value='regular')
    duration = reservations_projects.aggregate(duration=Sum(F('stop_time') - F('start_time')))['duration']
    result['all modes']['observing'] = duration.total_seconds() if duration else None
    # Non-production projects
    reservations_mixed = reservations.filter(project__cycles=cycle.pk).exclude(project__project_category__value='regular')
    duration = reservations_mixed.aggregate(duration=Sum(F('stop_time') - F('start_time')))['duration']
    result['all modes']['idle/test'] = duration.total_seconds() if duration else None

    # Get all the reservations in stand-alone mode
    reservations_sa = reservations.filter(specifications_doc__activity__type='stand-alone mode')
    duration = reservations_sa.aggregate(duration=Sum(F('stop_time') - F('start_time')))['duration']
    result['stand-alone mode']['total'] = duration.total_seconds() if duration else None
    #  No project
    duration = reservations_sa.filter(project=None).aggregate(duration=Sum(F('stop_time') - F('start_time')))['duration']
    result['stand-alone mode']['no project'] = duration.total_seconds() if duration else None
    # Production projects
    duration = (reservations_sa & reservations_projects).aggregate(duration=Sum(F('stop_time') - F('start_time')))['duration']
    result['stand-alone mode']['project'] = duration.total_seconds() if duration else None
    # Non-production projects
    duration = (reservations_sa & reservations_mixed).aggregate(duration=Sum(F('stop_time') - F('start_time')))['duration']
    result['stand-alone mode']['mixed/no project'] = duration.total_seconds() if duration else None

    # Get ILT durations from the previous durations
    result['ILT mode']['total'] = result['all modes']['total'] - result['stand-alone mode']['total'] if result['all modes']['total'] and result['stand-alone mode']['total'] else None
    result['ILT mode']['observing'] = result['all modes']['observing'] - result['stand-alone mode']['project'] if result['all modes']['observing'] and result['stand-alone mode']['project'] else None
    result['ILT mode']['idle/test'] = result['all modes']['idle/test'] - result['stand-alone mode']['mixed/no project'] if result['all modes']['idle/test'] and result['stand-alone mode']['mixed/no project'] else None

    return result


def _get_failures(cycle: models.Cycle, start: datetime, stop: datetime) -> {}:
    """
    Help function to retrieve failures rate info as a function of time.
    """
    from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess

    # TODO: See TMSS-662 for details.
    result = {'months': []}

    # Get SchedulingUnitBlueprints related to the cycle
    subs = models.SchedulingUnitBlueprint.objects.filter(draft__scheduling_set__project__cycles=cycle.pk)

    # Iterate through weeks and sum durations per week
    step, d = relativedelta(months=1), start
    while d < stop:
        total_per_month = 0
        total_failed_per_month = 0
        for sub in subs:
            # Aggregate total and successful durations
            if sub.observed_duration and d <= sub.observed_start_time < d + step and sub.observed_end_time < d + step:
                sup = SchedulingUnitProcess.objects.filter(su=sub).first()
                total_per_month += sub.observed_duration.total_seconds()
                total_failed_per_month += sub.observed_duration.total_seconds() if sup and sup.results_accepted is False else 0
        failed_perc = round(total_failed_per_month / total_per_month, 2) if total_per_month > 0 else None
        result['months'].append({'month': d.date().isoformat(), 'total': total_per_month, 'total_failed': total_failed_per_month, 'failed_perc': failed_perc})
        d += step

    return result


# Project Report

def create_project_report(request: Request, project: models.Project, start: datetime, stop: datetime) -> {}:
    """
    Create a project report as a JSON object.
    """
    with warnings.catch_warnings(record=True) as w:  # Catch potential TaskOverlappingWarning
        warnings.simplefilter('once', TaskOverlappingWarning)

        subs, durations = _get_subs_and_durations_from_project(project.pk, start, stop)
        result = {'project': project.pk, 'quota': _get_quotas_from_project(request, project.pk), 'SUBs': subs,
                  'durations': durations, 'LTA dataproducts': _get_lta_dataproducts(project.name, start, stop),
                  'SAPs exposure': _get_saps_exposure(project.pk, start, stop), 'processing_resources': _get_processing_resources(project.pk, start, stop)}
        result['contains_overlapping_observations'] = any([issubclass(warning.category, TaskOverlappingWarning) for warning in w])

    return result


def _get_quotas_from_project(request: Request, project_pk: int) -> []:
    """
    Help function to retrieve quotas.
    """
    project_quotas = models.ProjectQuota.objects.filter(project=project_pk)
    project_quotas_data = [serializers.ProjectQuotaSerializer(pq, context={'request': request}).data for pq in project_quotas]
    quotas = [{k: pqd[k] for k in ('id', 'resource_type_id', 'value')} for pqd in project_quotas_data]
    return quotas


def _get_subs_and_durations_from_project(project_pk: int, start: datetime, stop: datetime) -> ({}, {}):
    """
    Help function to retrieve durations and scheduling_units distinguished by success/fail.
    """
    from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess

    durations = {'total': 0, 'total_not_cancelled': 0, 'total_succeeded': 0, 'total_failed': 0,
                 'total_observed': 0, 'total_observed_succeeded': 0, 'total_observed_failed': 0}

    # Get SUBs related to the project
    project_subs = models.SchedulingUnitBlueprint.objects.filter(draft__scheduling_set__project__pk=project_pk)

    subs_succeeded, subs_failed = [], []
    for prio in models.PriorityQueueType.Choices:
        durations[f'total_succeeded_{prio.name}'], durations[f'total_observed_succeeded_{prio.name}'] = 0, 0
        subs = project_subs.filter(priority_queue=prio.value)
        for sub in subs:
            # Check time interval and go to the next iteration (SUB) if the check fails
            if (start and sub.scheduled_on_sky_start_time < start) or (stop and sub.scheduled_on_sky_stop_time > stop):
                continue
            # Gathering info for reporting
            sub_start_time = sub.scheduled_on_sky_start_time.isoformat() if sub.scheduled_on_sky_start_time else None
            sub_stop_time = sub.scheduled_on_sky_stop_time.isoformat() if sub.scheduled_on_sky_stop_time else None
            sub_duration = sub.duration.total_seconds()
            sub_observed_duration = sub.observed_duration.total_seconds() if sub.observed_duration else 0
            # Info about the SUB to be returned
            sub_info = {'id': sub.pk, 'name': sub.name, 'status': sub.status, 'start': sub_start_time,
                        'stop': sub_stop_time, 'duration': sub_duration, 'observed_duration': sub_observed_duration}
            try:
                sub_info['target'] = get_target_observation_task_name_from_specifications_doc(sub)
            except Exception as e:
                sub_info['target'] = None
            # Get SAS ID info
            subtasks = models.Subtask.objects.filter(task_blueprint__scheduling_unit_blueprint=sub)
            sub_info['SAS ID'] = {}
            for template_name in ('observation control', 'preprocessing pipeline', 'pulsar pipeline'):
                sub_info['SAS ID'][template_name] = [subtask.id for subtask in subtasks.filter(specifications_template__name=template_name)]
            # Get ingest info
            dataproducts = models.Dataproduct.objects.filter(producer__subtask__specifications_template__type='ingest').filter(producer__subtask__state__value='finished').filter(producer__subtask__task_blueprint__scheduling_unit_blueprint=sub)
            # TODO: Maybe it would be useful to implement an 'ingested_stop_time' (and even an 'ingested_start_time' and consequently an 'ingested_duration'?)
            sub_info['ingested_date'] = max(dataproducts, key=lambda x: x.producer.subtask.scheduled_on_sky_stop_time).producer.subtask.scheduled_on_sky_stop_time if dataproducts else None
            sub_info['ingested_data_size'] = dataproducts.aggregate(Sum('size'))['size__sum']

            # Gather durations
            durations['total'] += sub_duration  # Total
            if sub.status != 'cancelled':  # Not cancelled SUBs
                durations['total_not_cancelled'] += sub_duration
            durations['total_observed'] += sub_observed_duration  # Total observed
            # Distinguish successful and failed SUBs
            sup = SchedulingUnitProcess.objects.filter(su=sub).first()
            if sup and sup.results_accepted:    # Succeeded SUBs
                durations[f'total_succeeded_{prio.name}'] += sub_duration
                durations[f'total_observed_succeeded_{prio.name}'] += sub_observed_duration
                subs_succeeded.append(sub_info)
            elif sup and sup.results_accepted is False:  # Failed SUBs
                durations['total_failed'] += sub_duration
                durations['total_observed_failed'] += sub_observed_duration
                subs_failed.append(sub_info)
        # Gather prios durations as well
        durations['total_succeeded'] += durations[f'total_succeeded_{prio.name}']
        durations['total_observed_succeeded'] += durations[f'total_observed_succeeded_{prio.name}']

    subs = {'successful': subs_succeeded, 'failed': subs_failed}

    # Calculate percentages
    durations['not_cancelled_perc'] = round(durations['total_not_cancelled'] / durations['total'], 2) if durations['total'] > 0 else None
    durations['succeeded_perc'] = round(durations['total_succeeded'] / durations['total'], 2) if durations['total'] > 0 else None
    durations['failed_perc'] = round(durations['total_failed'] / durations['total'], 2) if durations['total'] > 0 else None
    durations['observed_perc'] = round(durations['total_observed'] / durations['total'], 2) if durations['total'] > 0 else None
    durations['observed_succeeded_perc'] = round(durations['total_observed_succeeded'] / durations['total'], 2) if durations['total'] > 0 else None
    durations['observed_failed_perc'] = round(durations['total_observed_failed'] / durations['total'], 2) if durations['total'] > 0 else None

    return subs, durations


def _get_lta_dataproducts(project_name: str, start: datetime, stop: datetime) -> {}:
    """
    Help function to retrieve the sum of the LTA dataproducts sizes.
    """
    # Query dataproducts from Subtasks of type 'ingest' within 'finished' status
    dataproducts = models.Dataproduct.objects.filter(producer__subtask__specifications_template__type='ingest') \
        .filter(producer__subtask__state__value='finished') \
        .filter(producer__subtask__task_blueprint__draft__scheduling_unit_draft__scheduling_set__project__name=project_name)
    # Filter basing on date interval if passed
    dataproducts = dataproducts.filter(producer__subtask__scheduled_on_sky_start_time__gte=start) if start else dataproducts
    dataproducts = dataproducts.filter(producer__subtask__scheduled_on_sky_stop_time__lte=stop) if stop else dataproducts

    return dataproducts.aggregate(Sum('size'))


def _get_saps_exposure(project_pk: int, start: datetime, stop: datetime) -> {}:
    """
    Help function to retrieve SAPs.
    """
    result = {}

    tasks = models.TaskBlueprint.objects.filter(scheduling_unit_blueprint__draft__scheduling_set__project=project_pk)
    for task in tasks:
        # Get subtasks and filter basing on date interval if passed
        subtasks = models.Subtask.objects.filter(task_blueprint=task, specifications_template__type='observation')
        subtasks = subtasks.filter(scheduled_on_sky_start_time__gte=start) if start else subtasks
        subtasks = subtasks.filter(scheduled_on_sky_stop_time__lte=stop) if stop else subtasks
        # Get SAPs
        saps = task.specifications_doc.get('SAPs', [])
        for sap in saps:
            sap_name = sap['name']
            result[sap_name] = 0
            # Aggregate durations per each SAP name
            for subtask in subtasks:
                digital_points = subtask.specifications_doc.get('stations', {}).get('digital_pointings', [])
                for dp in digital_points:
                    if dp['name'] == sap_name:
                        result[sap_name] += subtask.duration.total_seconds()

    return result


def _get_processing_resources(project_pk: int, start: datetime, stop: datetime) -> {}:
    """
    Help function to retrieve info about processing resources.
    """
    result = {'CPUtimeRAW': None, 'CPU_time_used': 0}

    # Get subtasks and filter basing on date interval if passed
    subtasks = models.Subtask.objects.filter(task_blueprint__scheduling_unit_blueprint__draft__scheduling_set__project=project_pk)
    subtasks = subtasks.filter(scheduled_on_sky_start_time__gte=start) if start else subtasks
    subtasks = subtasks.filter(scheduled_on_sky_stop_time__lte=stop) if stop else subtasks

    # TODO: Update this accordingly when we'll have feedbacks. See TMSS-592.
    pipeline_subtasks = subtasks.filter(specifications_template__type='pipeline')
    for pipeline in pipeline_subtasks:
        parallel_tasks = pipeline.specifications_doc.get('cluster_resources', {}).get('parallel_tasks', 0)
        cores_per_task = pipeline.specifications_doc.get('cluster_resources', {}).get('cores_per_task', 0)
        result['CPU_time_used'] += (pipeline.duration.total_seconds() * (parallel_tasks * cores_per_task) / 1000.0)

    return result
