"""
This file contains the database models
"""

import logging
logger = logging.getLogger(__name__)

from django.db import transaction
from django.db.models import Model, CharField, DateTimeField, BooleanField, ForeignKey, CASCADE, IntegerField, FloatField, SET_NULL, PROTECT, ManyToManyField, UniqueConstraint, QuerySet, OneToOneField, Q, Max
from django.contrib.postgres.fields import JSONField
from enum import Enum
from django.db.models.expressions import RawSQL
from django.db.models.deletion import ProtectedError
from .common import AbstractChoice, BasicCommon, Template, NamedCommon, NamedVersionedCommon, TemplateSchemaMixin, NamedCommonPK, RefreshFromDbInvalidatesCachedPropertiesMixin, ProjectPropertyMixin
from lofar.common.json_utils import validate_json_against_schema, validate_json_against_its_schema, add_defaults_to_json_object_for_schema
from lofar.sas.tmss.tmss.exceptions import *
from django.core.exceptions import ValidationError
import datetime
import warnings
from collections import Counter
from django.utils.functional import cached_property
from pprint import pformat
from lofar.sas.tmss.tmss.exceptions import TMSSException
from lofar.sas.tmss.tmss.exceptions import BlueprintCreationException, TMSSException
from django.db.models import Count
from copy import deepcopy
from lofar.common.util import subdict_of_pointer_items
from .permissions import TMSSUser as User

#
# I/O
#

# choices

class Role(AbstractChoice):
    """Defines the model and predefined list of possible Role's for TaskConnectorType.
    The items in the Choises class below are automagically populated into the database via a data migration.
    When changing/adding/removing items to the Choices, please update the common json schema for tasks as well."""
    class Choices(Enum):
        CORRELATOR = "correlator"
        BEAMFORMER = "beamformer"
        INSPECTION_PLOTS = "inspection plots"
        CALIBRATOR = "calibrator"
        TARGET = "target"
        ANY = "any"


class IOType(AbstractChoice):
    """Defines the model and predefined list of possible IOType's for TaskConnectorType.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        INPUT = "input"
        OUTPUT = "output"
        # maybe we can add an IN_PLACE="in_place" option in the future, but for now it's not needed.


class Datatype(AbstractChoice):
    """Defines the model and predefined list of possible Datatype's for TaskConnectorType.
    The items in the Choises class below are automagically populated into the database via a data migration.
    When changing/adding/removing items to the Choices, please update the common json schema for tasks as well."""
    class Choices(Enum):
        VISIBILITIES = "visibilities"
        TIME_SERIES = "time series"
        INSTRUMENT_MODEL = "instrument model"
        IMAGE = "image"
        QUALITY = "quality"
        PULSAR_PROFILE = "pulsar profile"


class Dataformat(AbstractChoice):
    """Defines the model and predefined list of possible Dataformat's for TaskConnectorType.
    The items in the Choises class below are automagically populated into the database via a data migration.
    When changing/adding/removing items to the Choices, please update the common json schema for tasks as well."""
    class Choices(Enum):
        MEASUREMENTSET = "MeasurementSet"
        BEAMFORMED = "Beamformed"
        QA_HDF5 = "QA_HDF5"
        QA_PLOTS = "QA_Plots"
        PULP_SUMMARY = "pulp summary"
        PULP_ANALYSIS = "pulp analysis"


class SchedulingRelationPlacement(AbstractChoice):
    """Defines the model and predefined list of possible Placements for Task Scheduling Relation.
        The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        AFTER = "after"
        BEFORE = "before"
        PARALLEL = "parallel"


class SystemSettingFlag(AbstractChoice):
    """Defines the model and predefined list of possible Flags to be used in Setting.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        DYNAMIC_SCHEDULING_ENABLED = "dynamic_scheduling_enabled"



class Quantity(AbstractChoice):
    """Defines the model and predefined list of possible period categories to be used in a.o. ProjectQuota and CycleQuota.
       Please note that, by agreement, all values of a certain quantity are stored in SI units.
       For example: a duration has the quantity "time" which is stored in "seconds".
       We only store the quatity-type in the database, because we known the correct SI units for each quantity, so there is no need to store the unit as well.
        The items in the Choices class below are automagically populated into the database via a data migration."""

    class Choices(Enum):
        # these are the basic quantities that we currently use. More can be added if needed.
        TIME = "time"
        BYTES = "bytes"
        NUMBER = "number"


class PeriodCategory(AbstractChoice):
    """Defines the model and predefined list of possible period categories to be used in Project as a policy for managing the project's lifetime.
        The items in the Choices class below are automagically populated into the database via a data migration."""

    class Choices(Enum):
        SINGLE_CYCLE = "single_cycle"
        LONG_TERM = "long_term"
        UNBOUNDED = "unbounded"


class ProjectCategory(AbstractChoice):
    """Defines the model and predefined list of possible project categories to be used in Project.
        The items in the Choices class below are automagically populated into the database via a data migration."""

    class Choices(Enum):
        REGULAR = "regular"
        USER_SHARED_SUPPORT = "user_shared_support"
        COMMISSIONING = "commissioning"
        DDT = "ddt"
        TEST = "test"


class TaskType(AbstractChoice):
    """Defines the model and predefined list of possible TaskType's for Task.
    The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        OBSERVATION = "observation"
        PIPELINE = "pipeline"
        INGEST = "ingest"
        CLEANUP = 'cleanup'
        MAINTENANCE = "maintenance"
        OTHER = 'other'


class PriorityQueueType(AbstractChoice):
    """Defines the possible priority queues for SchedulingUnits.
    The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        A = "A"
        B = "B"


class ProjectState(AbstractChoice):
    """Defines the model and predefined list of possible ProjectState's of a Project.
    The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        OPENED = "opened"
        ACTIVE = "active"
        FINISHED = "finished"
        CANCELLED = "cancelled"
        SUSPENDED = "suspended"


class SystemEventType(AbstractChoice):
    """Defines the model and predefined list of possible issue_types of a SystemEvent.
    The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        CEP = "cep"
        COBALT = "cobalt"
        ENVIRONMENT = "environment"
        HUMAN ="human"
        NETWORK = "network"
        STATION = "station"
        SYSTEM = "system"
        OTHER = "other"


class SystemEventSubtype(AbstractChoice):
    """Defines the model and predefined list of possible issue_subtypes of a SystemEvent.
    The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        CRASH = "crash"
        DATALOSS = "dataloss"
        HARDWARE = "hardware"
        NOISY = "noisy"
        OSCILLATING = "oscillating"
        RFI = "rfi"
        SETUP = "setup"
        TEMPERATURE = "temperature"
        OTHER = "other"


class SystemEventSeverity(AbstractChoice):
    """Defines the model and predefined list of possible severity levels of a SystemEvent.
    The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        FAILURE = "failure"
        MINOR = "minor"
        MAJOR = "major"


class SystemEventStatus(AbstractChoice):
    """Defines the model and predefined list of possible statuses of a SystemEvent.
    The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        OPEN = "open"
        ANALYSED = "analysed"


# concrete models

class Setting(BasicCommon):
    name = OneToOneField('SystemSettingFlag', null=False, on_delete=PROTECT, primary_key=True)
    value = BooleanField(null=False)


class TaskConnectorType(Model):
    ''' Describes the data type & format combinations a Task can accept or produce. The "role" is used to distinguish
        inputs (or outputs) that have the same data type & format, but are used in different ways by the task. For
        example, a calibration pipeline accepts measurement sets only, but distinghuishes between CALIBRATOR and
        TARGET roles.'''
    role = ForeignKey('Role', null=False, on_delete=PROTECT)
    datatype = ForeignKey('Datatype', null=False, on_delete=PROTECT)
    dataformat = ForeignKey('Dataformat', null=False, on_delete=PROTECT)
    task_template = ForeignKey("TaskTemplate", related_name='connector_types', null=False, on_delete=CASCADE)
    iotype = ForeignKey('IOType', null=False, on_delete=PROTECT, help_text="Is this connector an input or output")

    class Meta:
        constraints = [UniqueConstraint(fields=['role', 'datatype', 'dataformat', 'task_template', 'iotype'], name='task_connector_type_unique_combination')]


#
# Templates
#

class CommonSchemaTemplate(Template):
    '''A Template model for common (reusable) schema's'''
    pass


class BaseStrategyTemplate(NamedVersionedCommon):
    '''
    A (subclass of a) BaseStrategyTemplate is a template in the sense that contains a 'predefined' JSON document. It is not a JSON-schema, but a document!
    '''
    template = JSONField(null=False, help_text='JSON-data document compliant with the JSON-schema in the referenced template. '
                                               'This strategy template is like a predefined recipe with all the correct settings, and defines which parameters the user can alter.')

    class Meta(NamedVersionedCommon.Meta):
        abstract = True

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.auto_set_version_number()
        super().save(force_insert, force_update, using, update_fields)


class SchedulingUnitObservingStrategyTemplate(BaseStrategyTemplate):
    '''
    A SchedulingUnitObservingStrategyTemplate is a template in the sense that it serves as a template to fill in json data objects conform its referred scheduling_unit_template.
    It is however not derived from the (abstract) Template super-class, because the Template super class is for JSON schemas, not JSON data objects.
    '''
    scheduling_unit_template = ForeignKey("SchedulingUnitTemplate", on_delete=PROTECT, null=False, help_text="")

    class Meta(BaseStrategyTemplate.Meta):
        abstract = False

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.template and self.scheduling_unit_template_id and self.scheduling_unit_template.schema:
            try:
                # validate. Do not use a cache upon this save moment, so we validate against the latest remote schemas
                validate_json_against_schema(self.template, self.scheduling_unit_template.schema)
            except Exception as e:
                # log the error for debugging and re-raise
                logger.error("Error while validating SchedulingUnitObservingStrategyTemplate name='%s' id='%s' error: %s\ntemplate:\n%s",
                             self.name, self.id, e, pformat(self.template))
                raise

        super().save(force_insert, force_update, using, update_fields)

    @property
    def template_doc_complete_with_defaults(self) -> dict:
        template_doc = deepcopy(self.template)

        # loop over all tasks, and add the defaults to each task given the task's specifications_template
        for task_name, task_doc in list(template_doc.get('tasks',{}).items()):
            task_specifications_template = TaskTemplate.objects.get(name=task_doc['specifications_template']['name'], version=task_doc['specifications_template'].get('version',1))
            template_doc['tasks'][task_name] = add_defaults_to_json_object_for_schema(task_doc, task_specifications_template.schema,
                                                                                      cache=TemplateSchemaMixin._schema_cache, max_cache_age=TemplateSchemaMixin._MAX_SCHEMA_CACHE_AGE)

        # add the default constraints given the scheduling_constraints_template
        if 'scheduling_constraints_template' not in template_doc:
            try:
                # try to use a known default SchedulingConstraintsTemplate...
                constraints_template = SchedulingConstraintsTemplate.objects.get(name='constraints', version=1)
            except SchedulingConstraintsTemplate.DoesNotExist:
                # or use the latest if the default does not exist
                constraints_template = SchedulingConstraintsTemplate.objects.order_by('-updated_at').first()
                logger.info("SchedulingUnitObservingStrategyTemplate.template_doc_complete_with_defaults id=%s using SchedulingConstraintsTemplate name='%s' version=%s as default because none was specified.", self.id, constraints_template.name, constraints_template.version)

            # inject name/version into template_doc
            template_doc['scheduling_constraints_template'] = {"name": constraints_template.name, "version": constraints_template.version}

        constraints_template = SchedulingConstraintsTemplate.objects.get(name=template_doc['scheduling_constraints_template']['name'],
                                                                         version=template_doc['scheduling_constraints_template'].get('version',1))
        constraints_doc = add_defaults_to_json_object_for_schema(template_doc.get('scheduling_constraints_doc', {}), constraints_template.schema,
                                                                 cache=TemplateSchemaMixin._schema_cache, max_cache_age=TemplateSchemaMixin._MAX_SCHEMA_CACHE_AGE)
        template_doc['scheduling_constraints_doc'] = constraints_doc

        return template_doc

    @property
    def template_doc_with_just_the_parameters(self) -> dict:
        parameter_pointers = sum([p['refs'] for p in self.template.get('parameters',[])], [])

        template_doc = self.template_doc_complete_with_defaults
        template_doc = subdict_of_pointer_items(template_doc, parameter_pointers)

        return template_doc


class SchedulingSetStrategyTemplate(BaseStrategyTemplate):
    '''
    A SchedulingSetStrategyTemplate is a template in the sense that it serves as a template to fill in json data objects conform its referred scheduling_set_template.
    It is however not derived from the (abstract) Template super-class, because the Template super class is for JSON schemas, not JSON data objects.
    '''
    scheduling_set_template = ForeignKey("SchedulingSetTemplate", on_delete=PROTECT, null=False, help_text="")

    class Meta(NamedVersionedCommon.Meta):
        abstract = False

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.template and self.scheduling_set_template_id and self.scheduling_set_template.schema:
            try:
                # validate. Do not use a cache upon this save moment, so we validate against the latest remote schemas
                validate_json_against_schema(self.template, self.scheduling_set_template.schema)
            except Exception as e:
                # log the error for debugging and re-raise
                logger.exception("Error while validating SchedulingSetStrategyTemplate name='%s' id='%s' error: %s\ntemplate:\n%s",
                                 self.name, self.id, e, pformat(self.template))
                raise

        super().save(force_insert, force_update, using, update_fields)


class SchedulingSetTemplate(Template):
    pass


class SchedulingUnitTemplate(Template):
    pass


class SchedulingConstraintsTemplate(Template):
    pass


class TaskTemplate(Template):
    type = ForeignKey('TaskType', null=False, on_delete=PROTECT)

    @property
    def _need_to_auto_increment_version_number(self) -> bool:
        '''override, because for TaskTemplates we are interested in the template being used by other instances except for TaskConnectors'''
        if self.pk is None:
            # this is a new instance, so we need a new version number
            return True

        for rel_obj in self._meta.related_objects:
            if isinstance(rel_obj, TaskConnectorType):
                # skip TaskConnectorType, because each TaskTemplate is being used by TaskConnectorType which is not very interesting.
                # we're only interested in usage in mainly TaskDrafts
                continue
            if rel_obj.related_model.objects.filter(**{rel_obj.field.attname: self}).count() > 0:
                return True
        return False

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self._need_to_auto_increment_version_number and self.pk is not None:
            # this is an updated version of this template
            # as a kind gesture to the user we automatically copy all TaskConnectorTypes
            # from the current to the new version

            with transaction.atomic():
                # which connectors are currently referencing this template?
                current_connectors = self.connector_types.all()

                # do normal save, resulting in a new TaskTemplate instance
                super().save(force_insert, force_update, using, update_fields)

                # copy the current_connectors, and link them to this new template
                for connector in current_connectors:
                    TaskConnectorType.objects.create(task_template=self,
                                                     role=connector.role,
                                                     datatype=connector.datatype,
                                                     dataformat=connector.dataformat,
                                                     iotype=connector.iotype)
        else:
            # do normal save
            super().save(force_insert, force_update, using, update_fields)




class TaskRelationSelectionTemplate(Template):
    pass


class ReservationStrategyTemplate(BaseStrategyTemplate):
    '''
    A ReservationStrategyTemplate is a template in the sense that it serves as a template to fill in json data objects
    conform its referred reservation_template.
    It is however not derived from the (abstract) Template super-class, because the Template super class is for
    JSON schemas, not JSON data objects.
    '''
    reservation_template = ForeignKey("ReservationTemplate", on_delete=PROTECT, null=False, help_text="")

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.template and self.reservation_template_id and self.reservation_template.schema:
            # validate. Do not use a cache upon this save moment, so we validate against the latest remote schemas
            validate_json_against_schema(self.template, self.reservation_template.schema)

        super().save(force_insert, force_update, using, update_fields)


class ReservationTemplate(Template):
    pass

class SystemEventTemplate(Template):
    pass


#
# Instance Objects
#

class Cycle(RefreshFromDbInvalidatesCachedPropertiesMixin, NamedCommonPK):
    start = DateTimeField(help_text='Moment at which the cycle starts, that is, when its projects can run.')
    stop = DateTimeField(help_text='Moment at which the cycle officially ends.')

    @cached_property
    def duration(self) -> datetime.timedelta:
        '''the duration of the cycle (stop-start date)'''
        return self.stop - self.start

    # Project to Cycle references are ManyToMany now, so cannot be protected via on_delete on db level.
    # Instead, explicitly check whether there are projects linked to the Cycle and prevent delete
    # todo: verify correct behavior
    #  - Implemented: Cycles cannot be deleted if they have projects assigned to them
    #  - Also possible: Projects need at least one cycle
    def delete(self, *args, **kwargs):
        if len(self.projects.all()) > 0:
            raise ProtectedError(protected_objects=self.projects.all(), msg='This Cycle is referenced by %s project(s) and cannot be deleted.' % len(self.projects.all()))
        else:
            super().delete(*args, **kwargs)


class CycleQuota(Model):
    cycle = ForeignKey('Cycle', related_name="quota", on_delete=PROTECT, help_text='Cycle to which these quota apply.')
    value = FloatField(help_text='Resource Quota value')
    resource_type = ForeignKey('ResourceType', on_delete=PROTECT, help_text='Resource type.')


class Project(RefreshFromDbInvalidatesCachedPropertiesMixin, NamedCommonPK):
    # todo: cycles should be protected since we have to manually decide to clean up projects with a cycle or keep them without cycle, however, ManyToManyField does not allow for that
    cycles = ManyToManyField('Cycle', related_name='projects', blank=True, help_text='Cycles to which this project belongs (NULLable).')
    priority_rank = FloatField(null=False, help_text='Priority of this project w.r.t. other projects. Projects can interrupt observations of lower-priority projects.') # todo: add if needed: validators=[MinValueValidator(0.0), MaxValueValidator(1.0)]
    trigger_priority = IntegerField(default=1000, help_text='Priority of this project w.r.t. triggers.') # todo: verify meaning and add to help_text: "Triggers with higher priority than this threshold can interrupt observations of projects."
    auto_ingest = BooleanField(default=False, help_text='True if The data is ingested when the other scheduling unit tasks are finished. False if The data is ingested after approval in the QA validation workflow. At the end of this a flag is set that the data can be ingested.')
    can_trigger = BooleanField(default=False, help_text='True if this project is allowed to supply observation requests on the fly, possibly interrupting currently running observations (responsive telescope).')
    private_data = BooleanField(default=True, help_text='True if data of this project is sensitive. Sensitive data is not made public.')
    expert = BooleanField(default=False, help_text='Expert projects put more responsibility on the PI.')
    filler = BooleanField(default=False, help_text='Use this project to fill up idle telescope time.')
    project_category = ForeignKey('ProjectCategory', help_text='Category this project falls under.', null=True, on_delete=PROTECT)
    period_category = ForeignKey('PeriodCategory', help_text='Policy for managing the lifetime of this project.', null=True, on_delete=PROTECT)
    auto_pin = BooleanField(default=False, help_text='True if the output_pinned flag of tasks in this project should be set True on creation.')
    piggyback_allowed_tbb = BooleanField(default=True, help_text='Piggyback key for TBB.')
    piggyback_allowed_aartfaac = BooleanField(default=True, help_text='Piggyback key for AARTFAAC.')
    path_to_project = "project"
    project_state = ForeignKey('ProjectState', default=ProjectState.Choices.OPENED.value, on_delete=PROTECT, null=False, help_text='The state this project is in.')

    @cached_property
    def duration(self) -> datetime.timedelta:
        '''return the overall duration of all tasks of this scheduling unit
        '''
        return self.relative_stop_time - self.relative_start_time

    @cached_property
    def project(self):
        return self


class ProjectQuota(Model):
    project = ForeignKey('Project', related_name="quota", on_delete=PROTECT, help_text='Project to wich this quota belongs.')  # protected to avoid accidents
    value = FloatField(help_text='Resource Quota value')
    resource_type = ForeignKey('ResourceType', on_delete=PROTECT, help_text='Resource type.')  # protected to avoid accidents


class ProjectQuotaArchiveLocation(RefreshFromDbInvalidatesCachedPropertiesMixin, Model):
    project_quota = ForeignKey('ProjectQuota', null=False, related_name="project_quota_archive_location", on_delete=PROTECT, help_text='The ProjectQuota for this archive location')
    archive_location = ForeignKey('Filesystem', null=False, on_delete=PROTECT, help_text='Location of an archive LTA cluster.')

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.project_quota.resource_type.quantity.value != Quantity.Choices.BYTES.value:
            raise ValueError("A ProjectQuotaArchiveLocation should have its project_quota defined in a 'bytes' quantity")
        if not self.archive_location.cluster.archive_site:
            raise ValueError("The archive_location of a ProjectQuotaArchiveLocation should be an archive site")
        super().save(force_insert, force_update, using, update_fields)

    @cached_property
    def archive_subdirectory(self) -> str:
        '''return name of the subdirectory where this project is stored. By default and convention this the the lower_case project name'''
        return self.project_quota.project.name.lower()

    @cached_property
    def full_archive_uri(self) -> str:
        '''return full uri where this project is stored. By default and convention this the the lower_case project name'''
        return "%s/%s/" % (self.archive_location.directory.rstrip('/'), self.archive_subdirectory)


class ResourceType(NamedCommonPK):
    quantity = ForeignKey('Quantity', null=False, on_delete=PROTECT, help_text='The quantity of this resource type.')


class SchedulingSet(NamedCommon):
    project = ForeignKey('Project', related_name="scheduling_sets", on_delete=PROTECT, help_text='Project to which this scheduling set belongs.')  # protected to avoid accidents


class SchedulingUnitDraft(NamedCommon, TemplateSchemaMixin, ProjectPropertyMixin):
    scheduling_set = ForeignKey('SchedulingSet', related_name='scheduling_unit_drafts', on_delete=CASCADE, help_text='Set to which this scheduling unit draft belongs.')
    specifications_template = ForeignKey('SchedulingUnitTemplate', on_delete=CASCADE, help_text='Schema used for specifications_doc.')
    observation_strategy_template = ForeignKey('SchedulingUnitObservingStrategyTemplate', on_delete=PROTECT, null=True, help_text='Observation Strategy Template used to create the specifications_doc.')
    scheduling_constraints_doc = JSONField(help_text='Scheduling Constraints for this run.', null=True)
    scheduling_constraints_template = ForeignKey('SchedulingConstraintsTemplate', on_delete=CASCADE, null=True, help_text='Schema used for scheduling_constraints_doc.')
    ingest_permission_required = BooleanField(default=False, help_text='Explicit permission is needed before the task.')
    piggyback_allowed_tbb = BooleanField(help_text='Piggyback key for TBB.', null=True)
    piggyback_allowed_aartfaac = BooleanField(help_text='Piggyback key for AARTFAAC.', null=True)
    priority_rank = FloatField(null=False, default=0.0, help_text='Priority of this scheduling unit w.r.t. other scheduling units within the same queue and project.')
    priority_queue = ForeignKey('PriorityQueueType', null=False, on_delete=PROTECT, default="A", help_text='Priority queue of this scheduling unit. Queues provide a strict ordering between scheduling units.')
    interrupts_telescope = BooleanField(default=False, null=False, help_text='boolean (default FALSE), which indicates whether this observation was triggered (responsive telescope)')
    path_to_project = 'scheduling_set__project'

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.specifications_template_id and self.specifications_template.schema is not None:
            # If this scheduling unit was created from an observation_strategy_template,
            # then make sure that the observation_strategy_template validates against this unit's specifications_template.schema
            if self.observation_strategy_template_id and self.observation_strategy_template.template:
                validate_json_against_schema(self.observation_strategy_template.template, self.specifications_template.schema)

        # This code only happens if the objects is not in the database yet. self._state.adding is True creating
        if self._state.adding:
            if hasattr(self, 'scheduling_set') and self.scheduling_set.project.auto_ingest is False:
                # When project.auto_ingest=False, the scheduling units will be created with ingest_permission_required = True
                self.ingest_permission_required=True

            # Propagate project piggyback values as default for scheduling_unit_draft
            if self.piggyback_allowed_tbb is None and hasattr(self, 'scheduling_set'):
                self.piggyback_allowed_tbb = self.scheduling_set.project.piggyback_allowed_tbb
            if self.piggyback_allowed_aartfaac is None and hasattr(self, 'scheduling_set'):
                self.piggyback_allowed_aartfaac = self.scheduling_set.project.piggyback_allowed_aartfaac

        self.annotate_validate_add_defaults_to_doc_using_template('scheduling_constraints_doc', 'scheduling_constraints_template')

        super().save(force_insert, force_update, using, update_fields)

    @cached_property
    def duration(self) -> datetime.timedelta:
        '''return the overall duration of all tasks of this scheduling unit
        '''
        return self.relative_stop_time - self.relative_start_time

    @cached_property
    def relative_start_time(self) -> datetime.timedelta:
        '''return the earliest relative start time of all tasks of this scheduling unit
        '''
        task_drafts = list(self.task_drafts.all())
        if task_drafts:
            return min(task_drafts, key=lambda x: x.relative_start_time).relative_start_time
        else:
            return datetime.timedelta(seconds=0)

    @cached_property
    def relative_stop_time(self) -> datetime.timedelta:
        '''return the latest relative stop time of all tasks of this scheduling unit
        '''
        task_drafts = list(self.task_drafts.all())
        if task_drafts:
            return max(task_drafts, key=lambda x: x.relative_stop_time).relative_stop_time
        else:
            return datetime.timedelta(seconds=0)

    @cached_property
    def project(self) -> Project:
        '''return the related project of this task
        '''
        return self.scheduling_set.project

    @property
    def specifications_doc(self) -> dict:
        '''Get a generated json specifications_doc (as dict) reflecting this current scheduling unit's task graph and settings'''
        # start with an empty specifications_doc which adheres to the scheduling_unit_template's json schema
        specifications_doc = self.specifications_template.get_default_json_document_for_schema()

        # add all tasks
        for task in self.task_drafts.order_by('name').all():
            specifications_doc['tasks'][task.name] = {'description': task.description,
                                                      'specifications_doc': task.specifications_doc,
                                                      'specifications_template': {'name': task.specifications_template.name,
                                                                                  'version': task.specifications_template.version}}


        # add all task relations
        for task_rel in TaskRelationDraft.objects.filter(Q(producer__scheduling_unit_draft__id=self.id) | Q(consumer__scheduling_unit_draft__id=self.id)).order_by('producer__name', 'consumer__name').all():
            specifications_doc['task_relations'].append({'producer': task_rel.producer.name,
                                                         'consumer': task_rel.consumer.name,
                                                         'input': {'role': task_rel.input_role.role.value,
                                                                   'datatype': task_rel.input_role.datatype.value,
                                                                   'dataformat': task_rel.input_role.dataformat.value},
                                                         'output': {'role': task_rel.output_role.role.value,
                                                                    'datatype': task_rel.output_role.datatype.value,
                                                                    'dataformat': task_rel.output_role.dataformat.value},
                                                         'selection_doc': task_rel.selection_doc,
                                                         'selection_template': task_rel.selection_template.name})

        # add all task scheduling relations
        for task_sched_rel in TaskSchedulingRelationDraft.objects.filter(Q(first__scheduling_unit_draft__id=self.id) | Q(second__scheduling_unit_draft__id=self.id)).order_by('first__name', 'second__name').all():
            specifications_doc['task_scheduling_relations'].append({'first': task_sched_rel.first.name,
                                                                    'second': task_sched_rel.second.name,
                                                                    'placement': task_sched_rel.placement.value,
                                                                    'time_offset': task_sched_rel.time_offset})

        # copy the scheduling_constraints
        if self.scheduling_constraints_template:
            specifications_doc['scheduling_constraints_template'] = {'name': self.scheduling_constraints_template.name,
                                                                     'version': self.scheduling_constraints_template.version}
            specifications_doc['scheduling_constraints_doc'] = self.scheduling_constraints_doc or self.scheduling_constraints_template.get_default_json_document_for_schema()

        # if this scheduling unit was created from an observation_strategy_template, then copy the parameters list
        if self.observation_strategy_template and self.observation_strategy_template.template:
            specifications_doc['parameters'] = self.observation_strategy_template.template.get('parameters', [])

        # ensure that the generated specifications_doc is valid according to the schema (raises if not valid)
        self.specifications_template.validate_document(specifications_doc)

        return specifications_doc

    def create_observing_strategy_template(self, name: str):
        '''create a new observing_strategy_template containing the specifications_doc which defines this scheduling_unit.'''
        return SchedulingUnitObservingStrategyTemplate.objects.create(name=name,
                                                                      description=self.description,
                                                                      scheduling_unit_template=SchedulingUnitTemplate.objects.get(name='scheduling unit'),
                                                                      template=self.specifications_doc)


class SchedulingUnitBlueprint(ProjectPropertyMixin, TemplateSchemaMixin, NamedCommon):
    class Status(Enum):
        DEFINED = "defined"
        FINISHED = "finished"
        CANCELLED = "cancelled"
        ERROR = "error"
        OBSERVING = "observing"
        OBSERVED = "observed"
        PROCESSING = "processing"
        PROCESSED = "processed"
        INGESTING = "ingesting"
        SCHEDULED = "scheduled"
        SCHEDULABLE = "schedulable"
        UNSCHEDULABLE = "unschedulable"

    # todo: are many of these fields supposed to be immutable in the database?
    #  Or are we fine to just not allow most users to change them?
    ingest_permission_required = BooleanField(default=False, help_text='Explicit permission is needed before the task.')
    ingest_permission_granted_since = DateTimeField(auto_now_add=False, null=True, help_text='The moment when ingest permission was granted.')
    specifications_template = ForeignKey('SchedulingUnitTemplate', on_delete=CASCADE, help_text='Template with JSON schema used for the specifications_doc (IMMUTABLE).')
    draft = ForeignKey('SchedulingUnitDraft', related_name='scheduling_unit_blueprints', on_delete=PROTECT, help_text='Scheduling Unit Draft which this run instantiates.')
    output_pinned = BooleanField(default=False, help_text='boolean (default FALSE), which blocks deleting unpinned dataproducts. When toggled ON, backend must pick SUB up for deletion. It also must when dataproducts are unpinned.')
    piggyback_allowed_tbb = BooleanField(help_text='Piggyback key for TBB.', null=True)
    piggyback_allowed_aartfaac = BooleanField(help_text='Piggyback key for AARTFAAC.', null=True)
    priority_rank = FloatField(null=False, default=0.0, help_text='Priority of this scheduling unit w.r.t. other scheduling units within the same queue and project.')
    priority_queue = ForeignKey('PriorityQueueType', null=False, on_delete=PROTECT, default="A", help_text='Priority queue of this scheduling unit. Queues provide a strict ordering between scheduling units.')
    scheduling_constraints_doc = JSONField(help_text='Scheduling Constraints for this run.', null=True)
    scheduling_constraints_template = ForeignKey('SchedulingConstraintsTemplate', on_delete=CASCADE, null=True, help_text='Schema used for scheduling_constraints_doc.')
    interrupts_telescope = BooleanField(default=False, null=False, help_text='boolean (default FALSE), which indicates whether this observation was triggered (responsive telescope)')
    path_to_project = 'draft__scheduling_set__project'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # keep original scheduling constraints to detect changes on save
        # Note: we cannot use self.scheduling_constraints_doc here since that causes an infinite loop of update_from_db
        if 'scheduling_constraints_doc' in kwargs.keys():
            self.__original_scheduling_constraints_doc = kwargs['scheduling_constraints_doc']
        else:
            field_names = [f.name for f in self._meta.fields]
            if len(args) == len(field_names):
                self.__original_scheduling_constraints_doc = args[field_names.index('scheduling_constraints_doc')]
            else:
                self.__original_scheduling_constraints_doc = None
        self.__original_scheduling_constraints_template_id = self.scheduling_constraints_template_id

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self._state.adding:
            # On creation, propagate the following scheduling_unit_draft attributes as default for the new scheduling_unit_blueprint
            for copy_field in ['ingest_permission_required', 'piggyback_allowed_tbb', 'piggyback_allowed_aartfaac',
                               'scheduling_constraints_doc', 'scheduling_constraints_template', 'interrupts_telescope']:
                if hasattr(self, 'draft'):
                    setattr(self, copy_field, getattr(self.draft, copy_field))
        else:
            # On updates, prevent changing the scheduling constraints doc or template if we are past schedulable state
            # todo: This causes a ton of tests to fail, e.g. t_workflow_qaworkflow returns errors 422
            if self.status not in [SchedulingUnitBlueprint.Status.DEFINED.value, SchedulingUnitBlueprint.Status.SCHEDULABLE.value] and \
                    ((self.__original_scheduling_constraints_doc is not None and self.scheduling_constraints_doc != self.__original_scheduling_constraints_doc) or
                     self.scheduling_constraints_template_id != self.__original_scheduling_constraints_template_id):
                raise TMSSException('The scheduling constraints of SchedulingUnitBlueprint pk=%s status=%s cannot be updated since it is not in defined or schedulable state.' % (self.pk, self.status))

        # update the original constraints value for comparison on next save
        self.__original_scheduling_constraints_doc = self.scheduling_constraints_doc
        self.__original_scheduling_constraints_template_id = self.scheduling_constraints_template_id

        if self._state.adding and self.interrupts_telescope:
            if self.project.can_trigger:
                from lofar.sas.tmss.services.scheduling.constraints import can_run_after
                start_time = datetime.datetime.utcnow() + datetime.timedelta(minutes=3)
                if self.scheduling_constraints_template is None or can_run_after(self, start_time):
                    logger.info('Triggered obs name=%s can run after start_time=%s. The scheduler will pick this up and cancel ongoing observations if necessary.' % (self.name, start_time))
                else:
                    logger.info('Triggered obs name=%s cannot run after start_time=%s. Adding it for book-keeping, but it will be unschedulable.' % (self.name, start_time))
                    # todo: set to unschedulable? This is a derived state and we do not have subtasks at this point. We could check this in 'status' of course, but this check is quite costly...
            else:
                msg = 'Triggered obs name=%s is rejected because its project name=%s does not allow triggering.' % (self.name, self.project.name)
                logger.warning(msg)
                raise BlueprintCreationException(msg)

        super().save(force_insert, force_update, using, update_fields)

    @cached_property
    def duration(self) -> datetime.timedelta:
        '''return the overall on sky duration of all tasks of this scheduling unit
        '''
        if self.scheduled_on_sky_start_time is None or self.scheduled_on_sky_stop_time is None:
            return self.relative_stop_time - self.relative_start_time
        else:
            return self.scheduled_on_sky_stop_time - self.scheduled_on_sky_start_time   # <- todo: do we ever want this?

    @cached_property
    def relative_start_time(self) -> datetime.timedelta:
        '''return the earliest relative start time of all tasks of this scheduling unit
        '''
        task_blueprints = list(self.task_blueprints.all())
        if task_blueprints:
            return min(task_blueprints, key=lambda x: x.relative_start_time).relative_start_time
        else:
            return datetime.timedelta(seconds=0)

    @cached_property
    def relative_stop_time(self) -> datetime.timedelta:
        '''return the latest relative stop time of all tasks of this scheduling unit
        '''
        task_blueprints = list(self.task_blueprints.all())
        if task_blueprints:
            return max(task_blueprints, key=lambda x: x.relative_stop_time).relative_stop_time
        else:
            return datetime.timedelta(seconds=0)

    @cached_property
    def scheduled_on_sky_start_time(self) -> datetime or None:
        '''return the earliest scheduled on sky start time of all tasks of this scheduling unit
        '''
        tasks_with_start_time = list(filter(lambda x: x.scheduled_on_sky_start_time is not None, self.task_blueprints.all()))
        if tasks_with_start_time:
            return min(tasks_with_start_time, key=lambda x: x.scheduled_on_sky_start_time).scheduled_on_sky_start_time
        else:
            return None

    @cached_property
    def scheduled_on_sky_stop_time(self) -> datetime or None:
        '''return the latest scheduled on sky stop time of all tasks of this scheduling unit
        '''
        tasks_with_stop_time = list(filter(lambda x: x.scheduled_on_sky_stop_time is not None, self.task_blueprints.all()))
        if tasks_with_stop_time:
            return max(tasks_with_stop_time, key=lambda x: x.scheduled_on_sky_stop_time).scheduled_on_sky_stop_time
        else:
            return None

    @cached_property
    def scheduled_process_start_time(self) -> datetime or None:
        '''return the earliest scheduled process start time of all tasks of this scheduling unit
        '''
        tasks_with_start_time = list(filter(lambda x: x.scheduled_process_start_time is not None, self.task_blueprints.all()))
        if tasks_with_start_time:
            return min(tasks_with_start_time, key=lambda x: x.scheduled_process_start_time).scheduled_process_start_time
        else:
            return None

    @cached_property
    def scheduled_process_stop_time(self) -> datetime or None:
        '''return the latest scheduled process stop time of all tasks of this scheduling unit
        '''
        tasks_with_stop_time = list(filter(lambda x: x.scheduled_process_stop_time is not None, self.task_blueprints.all()))
        if tasks_with_stop_time:
            return max(tasks_with_stop_time, key=lambda x: x.scheduled_process_stop_time).scheduled_process_stop_time
        else:
            return None

    @cached_property
    def actual_on_sky_start_time(self) -> datetime or None:
        '''return the earliest actual on sky start time of all tasks of this scheduling unit
        '''
        tasks_with_start_time = list(filter(lambda x: x.actual_on_sky_start_time is not None, self.task_blueprints.all()))
        if tasks_with_start_time:
            return min(tasks_with_start_time, key=lambda x: x.actual_on_sky_start_time).actual_on_sky_start_time
        else:
            return None

    @cached_property
    def actual_on_sky_stop_time(self) -> datetime or None:
        '''return the latest actual on sky stop time of all tasks of this scheduling unit
        '''
        tasks_with_stop_time = list(filter(lambda x: x.actual_on_sky_stop_time is not None, self.task_blueprints.all()))
        if tasks_with_stop_time:
            return max(tasks_with_stop_time, key=lambda x: x.actual_on_sky_stop_time).actual_on_sky_stop_time
        else:
            return None

    @cached_property
    def actual_process_start_time(self) -> datetime or None:
        '''return the earliest actual process start time of all tasks of this scheduling unit
        '''
        tasks_with_start_time = list(filter(lambda x: x.actual_process_start_time is not None, self.task_blueprints.all()))
        if tasks_with_start_time:
            return min(tasks_with_start_time, key=lambda x: x.actual_process_start_time).actual_process_start_time
        else:
            return None

    @cached_property
    def actual_process_stop_time(self) -> datetime or None:
        '''return the latest actual process stop time of all tasks of this scheduling unit
        '''
        tasks_with_stop_time = list(filter(lambda x: x.actual_process_stop_time is not None, self.task_blueprints.all()))
        if tasks_with_stop_time:
            return max(tasks_with_stop_time, key=lambda x: x.actual_process_stop_time).actual_process_stop_time
        else:
            return None

    @cached_property
    def on_sky_start_time(self) -> datetime or None:
        '''return the earliest on sky start time of all the observation tasks of this scheduling unit.
        '''
        observation_tasks_with_start_time = list(filter(lambda x: (x.specifications_template.type.value == TaskType.Choices.OBSERVATION.value and x.on_sky_start_time is not None), self.task_blueprints.all()))
        if observation_tasks_with_start_time:
            return min(observation_tasks_with_start_time, key=lambda x: x.on_sky_start_time).on_sky_start_time
        else:
            return None

    @cached_property
    def on_sky_stop_time(self) -> datetime or None:
        '''return the latest on sky stop time of all the observation tasks of this scheduling unit.
        '''
        observation_tasks_with_stop_time = list(filter(lambda x: (x.specifications_template.type.value == TaskType.Choices.OBSERVATION.value and x.on_sky_stop_time is not None), self.task_blueprints.all()))
        if observation_tasks_with_stop_time:
            return max(observation_tasks_with_stop_time, key=lambda x: x.on_sky_stop_time).on_sky_stop_time
        else:
            return None

    @cached_property
    def process_start_time(self) -> datetime or None:
        '''return the earliest process start time of all the tasks of this scheduling unit.
        '''
        tasks_with_start_time = list(filter(lambda x: x.process_start_time is not None, self.task_blueprints.all()))
        if tasks_with_start_time:
            return min(tasks_with_start_time, key=lambda x: x.process_start_time).process_start_time
        else:
            return None

    @cached_property
    def process_stop_time(self) -> datetime or None:
        '''return the latest process stop time of all the tasks of this scheduling unit.
        '''
        tasks_with_stop_time = list(filter(lambda x: x.process_stop_time is not None, self.task_blueprints.all()))
        if tasks_with_stop_time:
            return max(tasks_with_stop_time, key=lambda x: x.process_stop_time).process_stop_time
        else:
            return None

    @property
    def observed_start_time(self) -> datetime or None:
        """
        return the earliest on sky start time of all (observation) tasks of this scheduling unit with the status observed/finished
        """
        observed_tasks = []
        for task in self.task_blueprints.all():
            if task.specifications_template.type.value == TaskType.Choices.OBSERVATION.value and \
                    (task.status == "observed" or task.status == "finished") and task.scheduled_on_sky_start_time is not None:
                observed_tasks.append(task)
        if observed_tasks:
            return min(observed_tasks, key=lambda x: x.scheduled_on_sky_start_time).scheduled_on_sky_start_time
        else:
            return None

    @property
    def observed_end_time(self) -> datetime or None:
        """
        return the latest on sky stop time of all (observation) tasks of this scheduling unit with the status observed/finished
        """
        observed_tasks = []
        for task in self.task_blueprints.all():
            if task.specifications_template.type.value == TaskType.Choices.OBSERVATION.value and\
                    (task.status == "observed" or task.status == "finished") and task.scheduled_on_sky_stop_time is not None:
                observed_tasks.append(task)
        if observed_tasks:
            return max(observed_tasks, key=lambda x: x.scheduled_on_sky_stop_time).scheduled_on_sky_stop_time
        else:
            return None

    @property
    def observed_duration(self) -> datetime or None:
        """
        return the overall observed duration of all (observation) tasks of this scheduling unit
        """
        if self.observed_start_time and self.observed_end_time:
            observed_tasks = []
            for task in self.task_blueprints.all():
                if task.specifications_template.type.value == TaskType.Choices.OBSERVATION.value and \
                        (task.status == "observed" or task.status == "finished"):
                    observed_tasks.append(task)
            # TODO: For now we just assume that tasks run subsequently. Handle overlapping as exception in future.
            # Check if there are any overlapping obs task
            for i, t1 in enumerate(observed_tasks):
                for t2 in observed_tasks[i + 1:]:
                    if t1.start_time < t2.start_time < t1.stop_time:    # Trigger an exception if any overlaps.
                        warnings.warn('There are at least two tasks overlapping: observed_duration is not reliable.', TaskOverlappingWarning)
            return self.observed_end_time - self.observed_start_time
        else:
            return None

    @property
    def specifications_doc(self) -> dict:
        '''Get a generated json specifications_doc (as dict) reflecting this current scheduling unit's task graph and settings'''
        # start with an empty specifications_doc which adheres to the scheduling_unit_template's json schema
        specifications_doc = self.specifications_template.get_default_json_document_for_schema()

        # add all tasks
        for task in self.task_blueprints.order_by('name').all():
            specifications_doc['tasks'][task.name] = {'description': task.description,
                                                      'specifications_doc': task.specifications_doc,
                                                      'specifications_template': {'name': task.specifications_template.name,
                                                                                  'version': task.specifications_template.version}}

        # add all task relations
        for task_rel in TaskRelationBlueprint.objects.filter(Q(producer__scheduling_unit_blueprint__id=self.id) | Q(consumer__scheduling_unit_blueprint__id=self.id)).order_by('producer__name', 'consumer__name').all():
            specifications_doc['task_relations'].append({'producer': task_rel.producer.name,
                                                         'consumer': task_rel.consumer.name,
                                                         'input': {'role': task_rel.input_role.role.value,
                                                                   'datatype': task_rel.input_role.datatype.value,
                                                                   'dataformat': task_rel.input_role.dataformat.value},
                                                         'output': {'role': task_rel.output_role.role.value,
                                                                    'datatype': task_rel.output_role.datatype.value,
                                                                    'dataformat': task_rel.output_role.dataformat.value},
                                                         'selection_doc': task_rel.selection_doc,
                                                         'selection_template': task_rel.selection_template.name})

        # add all task scheduling relations
        for task_sched_rel in TaskSchedulingRelationBlueprint.objects.filter(Q(first__scheduling_unit_blueprint__id=self.id) | Q(second__scheduling_unit_blueprint__id=self.id)).order_by('first__name', 'second__name').all():
            specifications_doc['task_scheduling_relations'].append({'first': task_sched_rel.first.name,
                                                                    'second': task_sched_rel.second.name,
                                                                    'placement': task_sched_rel.placement.value,
                                                                    'time_offset': task_sched_rel.time_offset})

        if self.scheduling_constraints_template:
            # copy the scheduling_constraints
            specifications_doc['scheduling_constraints_template'] = {'name': self.scheduling_constraints_template.name,
                                                                     'version': self.scheduling_constraints_template.version}
            specifications_doc['scheduling_constraints_doc'] = self.scheduling_constraints_doc or self.scheduling_constraints_template.get_default_json_document_for_schema()

        # if this scheduling unit was created from an observation_strategy_template, then copy the parameters list
        if self.draft.observation_strategy_template and self.draft.observation_strategy_template.template:
            specifications_doc['parameters'] = self.draft.observation_strategy_template.template.get('parameters', [])

        # ensure that the generated specifications_doc is valid according to the schema (raises if not valid)
        self.specifications_template.validate_document(specifications_doc)

        return specifications_doc

    @property
    def specifications_doc_including_copies_for_failed_tasks(self) -> dict:
        '''Same as self.specifications_doc but then including copies of tasks and relations for failed tasks, intended to be used to update the draft and (re)create the copies the failed tasks for a "rerun".'''

        # start with the normal exported specifications_doc reflecting the current graph.
        specifications_doc = self.specifications_doc

        # add copies for all failed tasks
        for task in self.task_blueprints.order_by('name').all():
            if task.status in (TaskBlueprint.Status.ERROR.value, TaskBlueprint.Status.CANCELLED.value) and not task.is_obsolete:
                copy_of_task_definition = deepcopy(specifications_doc['tasks'][task.name])
                copy_of_task_name = task.name + " (copy for rerun)"
                copy_of_task_definition['description'] += " (copy for rerun)"
                specifications_doc['tasks'][copy_of_task_name] = copy_of_task_definition

                # automatically add copies of the relation to other tasks
                for relation in list(specifications_doc['task_relations']):
                    for p_or_c in ('producer', 'consumer'):
                        if relation[p_or_c] == task.name:
                            copy_of_task_relation = deepcopy(relation)
                            copy_of_task_relation[p_or_c] = copy_of_task_name
                            specifications_doc['task_relations'].append(copy_of_task_relation)

                # automatically add copies of the scheduling relation to other tasks
                for relation in list(specifications_doc['task_scheduling_relations']):
                    for f_or_s in ('first', 'second'):
                        if relation[f_or_s] == task.name:
                            copy_of_task_scheduling_relation = deepcopy(relation)
                            copy_of_task_scheduling_relation[f_or_s] = copy_of_task_name
                            specifications_doc['task_scheduling_relations'].append(copy_of_task_scheduling_relation)

        # ensure that the generated specifications_doc is valid according to the schema (raises if not valid)
        self.specifications_template.validate_document(specifications_doc)

        return specifications_doc

    @property
    def status(self):
        """
        Return the schedulingunit blueprint status which is derived from the taskblueprint status (which is derived
        from the subtasks states)
        See https://support.astron.nl/confluence/display/TMSS/Specification+Flow#SpecificationFlow-TaskBlueprints
        The status is displayed as extra field in rest api of the schedulingunit
        """

        # Determine status per task_type (unfortunately did not manage with updatable view)
        status_overview_counter = Counter()
        status_overview_counter_per_type = {type.value: Counter() for type in TaskType.Choices}
        for tb in self.task_blueprints.select_related('specifications_template').all():
            if tb.obsolete_since is None:
                status_overview_counter[tb.status] += 1
                status_overview_counter_per_type[tb.specifications_template.type.value][tb.status] += 1

        # The actual determination of the SchedulingunitBlueprint status
        if not self._task_graph_instantiated():
            return SchedulingUnitBlueprint.Status.DEFINED.value
        elif self._all_task_finished(status_overview_counter):
            return SchedulingUnitBlueprint.Status.FINISHED.value
        elif self._any_task_cancelled(status_overview_counter):
            return SchedulingUnitBlueprint.Status.CANCELLED.value
        elif self._any_task_error(status_overview_counter):
            return SchedulingUnitBlueprint.Status.ERROR.value
        elif self._any_task_unschedulable(status_overview_counter):
            return SchedulingUnitBlueprint.Status.UNSCHEDULABLE.value
        elif self._any_task_started_observed_finished(status_overview_counter):
            if not self._all_observation_task_observed_finished(status_overview_counter_per_type):
                return SchedulingUnitBlueprint.Status.OBSERVING.value
            elif not self._any_processing_task_started_or_finished(status_overview_counter_per_type):
                return SchedulingUnitBlueprint.Status.OBSERVED.value
            elif not self._all_processing_tasks_and_observation_tasks_finished(status_overview_counter_per_type):
                return SchedulingUnitBlueprint.Status.PROCESSING.value
            elif not self._any_ingest_task_started(status_overview_counter_per_type):
                return SchedulingUnitBlueprint.Status.PROCESSED.value
            else:
                return SchedulingUnitBlueprint.Status.INGESTING.value
        elif self._any_task_scheduled(status_overview_counter):
            return SchedulingUnitBlueprint.Status.SCHEDULED.value
        return SchedulingUnitBlueprint.Status.SCHEDULABLE.value

    def _task_graph_instantiated(self):
        from .scheduling import Subtask # import here to prevent circular imports
        return self._get_total_nbr_tasks() > 0 and \
               Subtask.objects.filter(task_blueprint__scheduling_unit_blueprint=self).count() > 0

    def _all_task_finished(self, status_overview_counter):
        return status_overview_counter["finished"] == self._get_total_nbr_tasks()

    @staticmethod
    def _any_task_cancelled(status_overview_counter):
        return status_overview_counter["cancelled"] > 0

    @staticmethod
    def _any_task_error(status_overview_counter):
        return status_overview_counter["error"] > 0

    @staticmethod
    def _any_task_obsolete(status_overview_counter):
        return status_overview_counter["obsolete"] > 0

    @staticmethod
    def _any_task_unschedulable(status_overview_counter):
        return status_overview_counter["unschedulable"] > 0

    @staticmethod
    def _any_task_started_observed_finished(status_overview_counter):
        return (status_overview_counter["started"] + status_overview_counter["observed"] + status_overview_counter["finished"]) > 0

    @staticmethod
    def _any_task_scheduled(status_overview_counter):
        return status_overview_counter["scheduled"] > 0

    def _all_observation_task_observed_finished(self, status_overview_counter_per_type):
        total_nbr_observation_tasks = self._get_total_nbr_observation_tasks(status_overview_counter_per_type)
        return (status_overview_counter_per_type["observation"]["observed"] +
                status_overview_counter_per_type["observation"]["finished"]) == total_nbr_observation_tasks

    @staticmethod
    def _any_processing_task_started_or_finished(status_overview_counter_per_type):
        return status_overview_counter_per_type["pipeline"]["started"] + status_overview_counter_per_type["pipeline"]["finished"] > 0

    def _all_processing_tasks_and_observation_tasks_finished(self, status_overview_counter_per_type):
        total_nbr_observation_tasks = self._get_total_nbr_observation_tasks(status_overview_counter_per_type)
        total_nbr_processing_tasks = self._get_total_nbr_processing_tasks(status_overview_counter_per_type)
        return (status_overview_counter_per_type["pipeline"]["finished"] == total_nbr_processing_tasks and
                status_overview_counter_per_type["observation"]["finished"] == total_nbr_observation_tasks)

    @staticmethod
    def _any_ingest_task_started(status_overview_counter_per_type):
        return status_overview_counter_per_type["ingest"]["started"] > 0

    def _get_total_nbr_tasks(self):
        return self.task_blueprints.count()

    @staticmethod
    def _get_total_nbr_observation_tasks(status_overview_counter_per_type):
        return len(tuple(status_overview_counter_per_type[TaskType.Choices.OBSERVATION.value].elements()))

    @staticmethod
    def _get_total_nbr_processing_tasks(status_overview_counter_per_type):
        return len(tuple(status_overview_counter_per_type[TaskType.Choices.PIPELINE.value].elements()))

    @staticmethod
    def _get_total_nbr_ingest_tasks(status_overview_counter_per_type):
        return len(tuple(status_overview_counter_per_type[TaskType.Choices.INGEST.value].elements()))

    @cached_property
    def project(self) -> Project:
        '''return the related project of this task
        '''
        return self.draft.scheduling_set.project

    @property
    def flat_station_list(self):
        """
        Get a flat list of stations of the scheduling unit sorted by name
        """
        lst_stations = []
        for sublist in self._get_recursively(self.specifications_doc, "stations"):
            for item in sublist:
                lst_stations.append(item)
        return sorted(list(set(lst_stations)))

    @property
    def station_groups(self):
        """
        Get the station groups of the scheduling unit
        """
        lst_station_groups = []
        for sublist in self._get_recursively(self.specifications_doc, "station_groups"):
            for item in sublist:
                lst_station_groups.append(item)
        return lst_station_groups

    def _get_recursively(self, search_dict, field):
        """
        Takes a dict with nested lists and dicts, and searches all dicts for a key of the field provided.
        """
        fields_found = []

        for key, value in search_dict.items():

            if key == field:
                fields_found.append(value)

            elif isinstance(value, dict):
                results = self._get_recursively(value, field)
                for result in results:
                    fields_found.append(result)

            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        more_results = self._get_recursively(item, field)
                        for another_result in more_results:
                            fields_found.append(another_result)

        return fields_found

    @property
    def system_events(self):
        events = []
        for task in self.task_blueprints.all():
            events += task.system_events
        return set(events)


class TaskDraft(NamedCommon, TemplateSchemaMixin, ProjectPropertyMixin):
    short_description = CharField(max_length=32, help_text='A short description of this task, usually the name of the target and abbreviated task type.', blank=True, default="")
    specifications_doc = JSONField(help_text='Specifications for this task.')
    scheduling_unit_draft = ForeignKey('SchedulingUnitDraft', related_name='task_drafts', on_delete=CASCADE, help_text='Scheduling Unit draft to which this task draft belongs.')
    specifications_template = ForeignKey('TaskTemplate', on_delete=CASCADE, help_text='Schema used for specifications_doc.') # todo: 'schema'?
    output_pinned = BooleanField(default=False, help_text='True if the output of this task is pinned to disk, that is, forbidden to be removed.')
    path_to_project = 'scheduling_unit_draft__scheduling_set__project'

    class Meta(NamedCommon.Meta):
        # ensure there are no tasks with duplicate names within one scheduling_unit
        constraints = [UniqueConstraint(fields=['name', 'scheduling_unit_draft'], name='TaskDraft_unique_name_in_scheduling_unit')]

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.annotate_validate_add_defaults_to_doc_using_template('specifications_doc', 'specifications_template')
        if self._state.adding:  # True on create, False on update, needs to be checked before super().save()
            super().save(force_insert, force_update, using, update_fields)
            if self.scheduling_unit_draft.scheduling_set.project.auto_pin:
                self.output_pinned = True
                self.save()
        else:
            super().save(force_insert, force_update, using, update_fields)

    @cached_property
    def successors(self) -> QuerySet:
        '''return the connect successor taskdraft(s) as queryset (over which you can perform extended queries, or return via the serializers/viewsets)
           If you want the result, add .all() like so: my_taskdraft.successors.all()
        '''
        # see comment on Subtask for why this needs to be raw sql
        return TaskDraft.objects.filter(id__in=RawSQL("SELECT successor_task.id FROM tmssapp_taskdraft as successor_task\n"
                                                      "INNER JOIN tmssapp_taskrelationdraft as task_rel on task_rel.consumer_id = successor_task.id\n"
                                                      "WHERE task_rel.producer_id = %s", params=[self.id]))

    @cached_property
    def predecessors(self) -> QuerySet:
        '''return the connect predecessor taskdraft(s) as queryset (over which you can perform extended queries, or return via the serializers/viewsets)
        If you want the result, add .all() like so: my_taskdraft.predecessors.all()
        '''
        # see comment on Subtask for why this needs to be raw sql
        return TaskDraft.objects.filter(id__in=RawSQL("SELECT successor_task.id FROM tmssapp_taskdraft as successor_task\n"
                                                      "INNER JOIN tmssapp_taskrelationdraft as task_rel on task_rel.producer_id = successor_task.id\n"
                                                      "WHERE task_rel.consumer_id = %s", params=[self.id]))

    @cached_property
    def duration(self) -> datetime.timedelta:
         '''returns the overall duration of this task
         '''
         return self.relative_stop_time - self.relative_start_time

    @cached_property
    def relative_start_time(self) -> datetime.timedelta:
        '''return the earliest relative start time of all subtasks of this task
        '''
        scheduling_relations = list(self.first_scheduling_relation.all()) + list(self.second_scheduling_relation.all())
        for scheduling_relation in scheduling_relations:
            # sometimes self._id does not exist so use self.id instead to avoid Exception
            if hasattr(self, '_id'):
                id = self._id
            else:
                id = self.id
            if scheduling_relation.first.id == id and scheduling_relation.placement_id == "after":
                previous_related_task_draft = TaskDraft.objects.get(id=scheduling_relation.second.id)
                time_offset = scheduling_relation.time_offset
                # todo: max of several relations
                if previous_related_task_draft.relative_stop_time:
                    return previous_related_task_draft.relative_stop_time + datetime.timedelta(seconds=time_offset)

            if scheduling_relation.second.id == id and scheduling_relation.placement_id == "before":
                previous_related_task_draft = TaskDraft.objects.get(id=scheduling_relation.first.id)
                time_offset = scheduling_relation.time_offset
                # todo: max of several relations
                if previous_related_task_draft.relative_stop_time:
                    return previous_related_task_draft.relative_stop_time + datetime.timedelta(seconds=time_offset)
        return datetime.timedelta(seconds=0)

    @cached_property
    def relative_stop_time(self) -> datetime.timedelta:
        '''return the latest relative stop time of all subtasks of this task
        '''
        # todo: when it was added, check if self.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
        try:
            duration = self.specifications_doc["duration"]
            return self.relative_start_time + datetime.timedelta(seconds=duration)
        except:
            pass
        return self.relative_start_time


    # JK, 28/07/20: After discussion with Sander, we probably only want the
    #  - duration on the scheduling_unit draft (based on relative start/stop times)
    #  - duration plus relative start/stop on the task draft.
    #  This provides an estimate of what is currently planned out in the draft, but does not confuse with timestamps of actual start/stop of the blueprints.
    #  Only on the blueprints, we also aggregate start_stop times as they are in the system
    #  I'll leave these code bits here for now, until we made up our minds about this, but this can probably be removed
    #
    # @cached_property
    # def duration(self) -> datetime.timedelta:
    #     '''returns the overall duration in seconds of all blueprints of this task
    #     # todo: is this the wanted behavior? Do you want to consider all the blueprints created from your draft or do you want to preview a new blueprint?
    #     '''
    #     if self.start_time is None or self.stop_time is None:
    #         # todo: calculate?
    #         return None
    #     else:
    #         return self.stop_time - self.start_time
    #
    # @cached_property
    # def start_time(self) -> datetime or None:
    #     '''return the earliest start time of all blueprints of this task
    #     # todo: is this the wanted behavior? Do you want to consider all the blueprints created from your draft or do you want to preview a new blueprint?
    #     '''
    #     blueprints_with_start_time = list(filter(lambda x: x.start_time is not None, self.task_blueprints.all()))
    #     if blueprints_with_start_time:
    #         return min(blueprints_with_start_time, key=lambda x: x.start_time).start_time
    #     else:
    #         # todo: calculate?
    #         return None
    #
    # @cached_property
    # def stop_time(self) -> datetime or None:
    #     '''return the latest stop time of all blueprints of this task
    #     # todo: is this the wanted behavior? Do you want to consider all the blueprints created from your draft or do you want to preview a new blueprint?
    #     '''
    #     blueprints_with_stop_time = list(filter(lambda x: x.stop_time is not None, self.task_blueprints.all()))
    #     if blueprints_with_stop_time:
    #         return max(blueprints_with_stop_time, key=lambda x: x.stop_time).stop_time
    #     else:
    #         # todo: calculate?
    #         return None


class TaskBlueprint(ProjectPropertyMixin, TemplateSchemaMixin, RefreshFromDbInvalidatesCachedPropertiesMixin, NamedCommon):
    class Status(Enum):
        DEFINED = "defined"
        FINISHED = "finished"
        CANCELLED = "cancelled"
        ERROR = "error"
        OBSERVED = "observed"
        STARTED = "started"
        SCHEDULED = "scheduled"
        SCHEDULABLE = "schedulable"
        UNSCHEDULABLE = "unschedulable"

    short_description = CharField(max_length=32, help_text='A short description of this task, usually the name of the target and abbreviated task type.', blank=True, default="")
    specifications_doc = JSONField(help_text='Schedulings for this task (IMMUTABLE).')
    specifications_template = ForeignKey('TaskTemplate', on_delete=CASCADE, help_text='Schema used for specifications_doc (IMMUTABLE).')
    draft = ForeignKey('TaskDraft', related_name='task_blueprints', null=True, on_delete=SET_NULL, help_text='TaskDraft from which this TaskBlueprint was created. If the TaskDraft is deleted the we loose this reference, which is ok.')
    scheduling_unit_blueprint = ForeignKey('SchedulingUnitBlueprint', related_name='task_blueprints', on_delete=CASCADE, help_text='Scheduling Unit Blueprint to which this task belongs.')
    output_pinned = BooleanField(default=False, help_text='True if the output of this task is pinned to disk, that is, forbidden to be removed.')
    path_to_project = 'draft__scheduling_unit_draft__scheduling_set__project'

    class Meta(NamedCommon.Meta):
        # ensure there are no tasks with duplicate names within one scheduling_unit,
        constraints = [UniqueConstraint(fields=['name', 'scheduling_unit_blueprint'], name='TaskBlueprint_unique_name_in_scheduling_unit')]

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.annotate_validate_add_defaults_to_doc_using_template('specifications_doc', 'specifications_template')
        super().save(force_insert, force_update, using, update_fields)

    @cached_property
    def successors(self) -> QuerySet:
        '''return the connect successor taskblueprint(s) as queryset (over which you can perform extended queries, or return via the serializers/viewsets)
           If you want the result, add .all() like so: my_taskblueprint.successors.all()
        '''
        # see comment on Subtask for why this needs to be raw sql
        return TaskBlueprint.objects.filter(id__in=RawSQL("SELECT successor_task.id FROM tmssapp_taskblueprint as successor_task\n"
                                                          "INNER JOIN tmssapp_taskrelationblueprint as task_rel on task_rel.consumer_id = successor_task.id\n"
                                                          "WHERE task_rel.producer_id = %s", params=[self.id]))

    @cached_property
    def predecessors(self) -> QuerySet:
        '''return the connect predecessor taskblueprint(s) as queryset (over which you can perform extended queries, or return via the serializers/viewsets)
        If you want the result, add .all() like so: my_taskblueprint.predecessors.all()
        '''
        # see comment on Subtask for why this needs to be raw sql
        return TaskBlueprint.objects.filter(id__in=RawSQL("SELECT predecessor_task.id FROM tmssapp_taskblueprint as predecessor_task\n"
                                                          "INNER JOIN tmssapp_taskrelationblueprint as task_rel on task_rel.producer_id = predecessor_task.id\n"
                                                          "WHERE task_rel.consumer_id = %s", params=[self.id]))

    @cached_property
    def duration(self) -> datetime.timedelta:
        '''return the overall duration of this task
        '''
        if self.scheduled_on_sky_start_time is None or self.scheduled_on_sky_stop_time is None:
            return self.relative_stop_time - self.relative_start_time
        else:
            return self.scheduled_on_sky_stop_time - self.schedule_on_sky_start_time

    @cached_property
    def relative_start_time(self) -> datetime.timedelta:
        '''The relative start time is relative to the start time of the 'start' of the parent scheduling unit.
        It's based on the scheduling_relation's from the scheduling_unit's specification,
        and hence it determines the order in which the tasks within the unit should be executed.
        '''
        scheduling_relations = list(self.first_scheduling_relation.all()) + list(self.second_scheduling_relation.all())
        for scheduling_relation in scheduling_relations:
            # sometimes self._id does not exist so use self.id instead to avoid Exception
            if hasattr(self, '_id'):
                id = self._id
            else:
                id = self.id
            if scheduling_relation.first.id == id and scheduling_relation.placement_id == "after":   # self.id and placement.value will hit the db, this does not
                    previous_related_task_blueprint = TaskBlueprint.objects.get(id=scheduling_relation.second.id)
                    time_offset = scheduling_relation.time_offset
                    # todo: max of several relations
                    if previous_related_task_blueprint.relative_stop_time:
                        return previous_related_task_blueprint.relative_stop_time + datetime.timedelta(seconds=time_offset)

            if scheduling_relation.second.id == id and scheduling_relation.placement_id == "before":   # self.id and placement.value will hit the db, this does not
                    previous_related_task_blueprint = TaskBlueprint.objects.get(id=scheduling_relation.first.id)
                    time_offset = scheduling_relation.time_offset
                    # todo: max of several relations
                    if previous_related_task_blueprint.relative_stop_time:
                        return previous_related_task_blueprint.relative_stop_time + datetime.timedelta(seconds=time_offset)
        return datetime.timedelta(seconds=0)

    @cached_property
    def relative_stop_time(self) -> datetime.timedelta:
        '''The relative_stop_time is the relative_start_time+duration.
        See relative_start_time for an explanation of it's intended usage.
        '''
        # todo: when it was added, check if subtask.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
        try:
            duration = datetime.timedelta(seconds=self.specifications_doc.get('duration', self.specifications_doc.get('target', {}).get('duration', 0)))
            return self.relative_start_time + duration
        except:
            pass
        return self.relative_start_time

    @cached_property
    def scheduled_on_sky_start_time(self) -> datetime or None:
        '''return the earliest scheduled on sky start time of all subtasks of this task
        '''
        subtasks_with_start_time = list(filter(lambda x: x.scheduled_on_sky_start_time is not None, self.subtasks.all()))
        if subtasks_with_start_time:
            return min(subtasks_with_start_time, key=lambda x: x.scheduled_on_sky_start_time).scheduled_on_sky_start_time
        else:
            return None

    @cached_property
    def scheduled_on_sky_stop_time(self) -> datetime or None:
        '''return the latest scheduled on sky stop time of all subtasks of this task
        '''
        subtasks_with_stop_time = list(filter(lambda x: x.scheduled_on_sky_stop_time is not None, self.subtasks.all()))
        if subtasks_with_stop_time:
            return max(subtasks_with_stop_time, key=lambda x: x.scheduled_on_sky_stop_time).scheduled_on_sky_stop_time
        else:
            return None

    @cached_property
    def scheduled_process_start_time(self) -> datetime or None:
        '''return the earliest scheduled process start time of all subtasks of this task
        '''
        subtasks_with_start_time = list(filter(lambda x: x.scheduled_process_start_time is not None, self.subtasks.all()))
        if subtasks_with_start_time:
            return min(subtasks_with_start_time, key=lambda x: x.scheduled_process_start_time).scheduled_process_start_time
        else:
            return None

    @cached_property
    def scheduled_process_stop_time(self) -> datetime or None:
        '''return the latest scheduled process stop time of all subtasks of this task
        '''
        subtasks_with_stop_time = list(filter(lambda x: x.scheduled_process_stop_time is not None, self.subtasks.all()))
        if subtasks_with_stop_time:
            return max(subtasks_with_stop_time, key=lambda x: x.scheduled_process_stop_time).scheduled_process_stop_time
        else:
            return None

    @cached_property
    def actual_on_sky_start_time(self) -> datetime or None:
        '''return the earliest actual on sky start time of all subtasks of this task
        '''
        subtasks_with_start_time = list(filter(lambda x: x.actual_on_sky_start_time is not None, self.subtasks.all()))
        if subtasks_with_start_time:
            return min(subtasks_with_start_time, key=lambda x: x.actual_on_sky_start_time).actual_on_sky_start_time
        else:
            return None

    @cached_property
    def actual_on_sky_stop_time(self) -> datetime or None:
        '''return the latest actual on sky stop time of all subtasks of this task
        '''
        subtasks_with_stop_time = list(filter(lambda x: x.actual_on_sky_stop_time is not None, self.subtasks.all()))
        if subtasks_with_stop_time:
            return max(subtasks_with_stop_time, key=lambda x: x.actual_on_sky_stop_time).actual_on_sky_stop_time
        else:
            return None

    @cached_property
    def actual_process_start_time(self) -> datetime or None:
        '''return the earliest actual process start time of all subtasks of this task
        '''
        subtasks_with_start_time = list(filter(lambda x: x.actual_process_start_time is not None, self.subtasks.all()))
        if subtasks_with_start_time:
            return min(subtasks_with_start_time, key=lambda x: x.actual_process_start_time).actual_process_start_time
        else:
            return None

    @cached_property
    def actual_process_stop_time(self) -> datetime or None:
        '''return the latest actual process stop time of all subtasks of this task
        '''
        subtasks_with_stop_time = list(filter(lambda x: x.actual_process_stop_time is not None, self.subtasks.all()))
        if subtasks_with_stop_time:
            return max(subtasks_with_stop_time, key=lambda x: x.actual_process_stop_time).actual_process_stop_time
        else:
            return None

    @cached_property
    def on_sky_start_time(self) -> datetime or None:
        '''return the earliest on sky start time of all the observation subtasks of this task.
        '''
        observation_control_subtasks = list(filter(lambda x: (x.specifications_template.name == 'observation control' and x.on_sky_start_time is not None), self.subtasks.all()))
        if observation_control_subtasks:
            return min(observation_control_subtasks, key=lambda x: x.on_sky_start_time).on_sky_start_time
        else:
            return None

    @cached_property
    def on_sky_stop_time(self) -> datetime or None:
        '''return the latest on sky stop time of all the observation subtasks of this task.
        '''
        observation_control_subtasks = list(filter(lambda x: (x.specifications_template.name == 'observation control' and x.on_sky_stop_time is not None), self.subtasks.all()))
        if observation_control_subtasks:
            return max(observation_control_subtasks, key=lambda x: x.on_sky_stop_time).on_sky_stop_time
        else:
            return None

    @cached_property
    def process_start_time(self) -> datetime or None:
        '''return the earliest process start time of all the subtasks of this task.
        '''
        subtasks_with_start_time = list(filter(lambda x: x.process_start_time is not None, self.subtasks.all()))
        if subtasks_with_start_time:
            return min(subtasks_with_start_time, key=lambda x: x.process_start_time).process_start_time
        else:
            return None

    @cached_property
    def process_stop_time(self) -> datetime or None:
        '''return the latest process stop time of all the subtasks of this task.
        '''
        subtasks_with_stop_time = list(filter(lambda x: x.process_stop_time is not None, self.subtasks.all()))
        if subtasks_with_stop_time:
            return max(subtasks_with_stop_time, key=lambda x: x.process_stop_time).process_stop_time
        else:
            return None

    @cached_property
    def is_obsolete(self) -> bool:
        '''is this task obsolete (because any of its subtasks are obsolete)'''
        return self.obsolete_since is not None

    @cached_property
    def obsolete_since(self) -> datetime or None:
        '''return the earliest obsolete_since time of all subtasks of this task
        '''
        if self.subtasks.filter(obsolete_since__isnull=True).exists():
            # there is at least one subtask that is not obsolete, so this task is not obsolete as well
            return None

        # return latest subtask obsolete_since timestamp, since this is when this task got obsolete
        return self.subtasks.filter(obsolete_since__isnull=False).values('obsolete_since').aggregate(Max('obsolete_since'))['obsolete_since__max']

    @property
    def status(self):
        """
        Return the taskblueprint status which is derived from the subtasks status
        See https://support.astron.nl/confluence/display/TMSS/Specification+Flow#SpecificationFlow-TaskBlueprints
        The status is displayed as extra field in rest api of the taskblueprint
        """

        # one call to the db fetching only the needed values
        # we ignore non-primary subtasks that have been marked obsolete
        subtasks = tuple(self.subtasks.exclude(primary=False, obsolete_since__isnull=False).values('state', 'specifications_template__type_id').all())

        # simple python filtering counting by state and type. Saves many round-trip-db calls!
        nr_of_subtasks = len(subtasks)
        if nr_of_subtasks == 0:
            return TaskBlueprint.Status.DEFINED.value

        if any(s for s in subtasks if s['state'] == 'defining'):
            return TaskBlueprint.Status.DEFINED.value

        if len([s for s in subtasks if s['state'] == 'finished']) == nr_of_subtasks:
            return TaskBlueprint.Status.FINISHED.value

        if any(s for s in subtasks if s['state'] in ('cancelling', 'cancelled')):
            return TaskBlueprint.Status.CANCELLED.value

        if any(s for s in subtasks if s['state'] == 'error'):
            return TaskBlueprint.Status.ERROR.value

        observations = [s for s in subtasks if s['specifications_template__type_id'] == 'observation']
        if observations and all(obs and obs['state'] in ('finishing', 'finished') for obs in observations):
            return TaskBlueprint.Status.OBSERVED.value

        if any(s for s in subtasks if s['state'] in ('starting','started','queueing','queued','finishing','finished')):
            return TaskBlueprint.Status.STARTED.value

        if any(s for s in subtasks if s['state'] == 'scheduled'):
            return TaskBlueprint.Status.SCHEDULED.value

        if any(s for s in subtasks if s['state'] == 'unschedulable'):
            return TaskBlueprint.Status.UNSCHEDULABLE.value

        return TaskBlueprint.Status.SCHEDULABLE.value


class TaskRelationDraft(BasicCommon, TemplateSchemaMixin, ProjectPropertyMixin):
    selection_doc = JSONField(help_text='Filter for selecting dataproducts from the output role.')
    selection_template = ForeignKey('TaskRelationSelectionTemplate', on_delete=CASCADE, help_text='Schema used for selection_doc.')

    # caveat: it might look like producer has an incorrect related_name='consumed_by'. But it really is correct, denends on the way you look at it
    producer = ForeignKey('TaskDraft', related_name='consumed_by', on_delete=CASCADE, help_text='Task Draft that has the output connector. NOTE: The producer does typically, but not necessarily, belong to the same Scheduling Unit (or even the same Project) as the consumer.')
    # caveat: it might look like consumer has an incorrect related_name='produced_by'. But it really is correct, denends on the way you look at it
    consumer = ForeignKey('TaskDraft', related_name='produced_by', on_delete=CASCADE, help_text='Task Draft that has the input connector.')

    # this relation descibes a transfer of data from the output_role of the producer to the input_role of the consumer
    input_role = ForeignKey('TaskConnectorType', related_name='taskrelationdraft_input_roles', on_delete=CASCADE, help_text='Input connector type (what kind of data is given to the consumer).')
    output_role = ForeignKey('TaskConnectorType', related_name='taskrelationdraft_output_roles', on_delete=CASCADE, help_text='Output connector type (what kind of data is taken from the producer).')
    path_to_project = 'producer__scheduling_unit_draft__scheduling_set__project'

    class Meta(BasicCommon.Meta):
        # ensure there are no duplicate relations between tasks with the same in/out roles.
        constraints = [UniqueConstraint(fields=['producer', 'consumer', 'input_role', 'output_role'], name='TaskRelationDraft_unique_relation')]

        # ensure that the roles are compatible, that is, the output we take is suitable for the input we provide to:
        # input_role.dataformat == output_role.dataformat
        # input_role.datatype == outputrole.datatype
        # input_role.output = False
        # output_role.output = True

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.annotate_validate_add_defaults_to_doc_using_template('selection_doc', 'selection_template')
        super().save(force_insert, force_update, using, update_fields)


class TaskRelationBlueprint(BasicCommon, TemplateSchemaMixin, ProjectPropertyMixin):
    selection_doc = JSONField(help_text='Filter for selecting dataproducts from the output role.')

    # caveat: it might look like producer has an incorrect related_name='consumed_by'. But it really is correct, denends on the way you look at it
    producer = ForeignKey('TaskBlueprint', related_name='consumed_by', on_delete=CASCADE, help_text='Task Blueprint that has the output connector.')
    # caveat: it might look like consumer has an incorrect related_name='produced_by'. But it really is correct, denends on the way you look at it
    consumer = ForeignKey('TaskBlueprint', related_name='produced_by', on_delete=CASCADE, help_text='Task Blueprint that has the input connector.')

    input_role = ForeignKey('TaskConnectorType', related_name='taskrelationblueprint_input_roles', on_delete=CASCADE, help_text='Input connector type (what kind of data can be taken as input).')
    output_role = ForeignKey('TaskConnectorType', related_name='taskrelationblueprint_output_roles', on_delete=CASCADE, help_text='Output connector type (what kind of data can be created as output).')
    draft = ForeignKey('TaskRelationDraft', null=True, on_delete=SET_NULL, related_name='blueprints', help_text='TaskRelationDraft from which this TaskRelationBlueprint was created. If the TaskRelationDraft is deleted the we loose this reference, which is ok.')
    selection_template = ForeignKey('TaskRelationSelectionTemplate', on_delete=CASCADE, help_text='Schema used for selection_doc.')  # todo: 'schema'?
    path_to_project = 'draft__producer__scheduling_unit_draft__scheduling_set__project'

    class Meta(BasicCommon.Meta):
        # ensure there are no duplicate relations between tasks with the same in/out roles.
        constraints = [UniqueConstraint(fields=['producer', 'consumer', 'input_role', 'output_role'], name='TaskRelationBlueprint_unique_relation')]

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.annotate_validate_add_defaults_to_doc_using_template('selection_doc', 'selection_template')
        super().save(force_insert, force_update, using, update_fields)


class TaskSchedulingRelationBlueprint(BasicCommon, ProjectPropertyMixin):
    draft = ForeignKey('TaskSchedulingRelationDraft', null=True, on_delete=SET_NULL, related_name='blueprints', help_text='TaskSchedulingRelationDraft from which this TaskSchedulingRelationBlueprint was created. If the TaskSchedulingRelationDraft is deleted the we loose this reference, which is ok.')
    first = ForeignKey('TaskBlueprint', related_name='first_scheduling_relation', on_delete=CASCADE, help_text='First Task Blueprint to connect.')
    second = ForeignKey('TaskBlueprint', related_name='second_scheduling_relation', on_delete=CASCADE, help_text='Second Task Blueprint to connect.')
    placement = ForeignKey('SchedulingRelationPlacement', null=False, default="after", on_delete=PROTECT, help_text='Task scheduling relation placement.')
    time_offset = IntegerField(default=60, help_text='Time offset of start of second task with respect to start of first task.')  
    path_to_project = 'first__draft__scheduling_unit_draft__scheduling_set__project'

    class Meta(BasicCommon.Meta):
        # ensure there are no duplicate scheduling relations between tasks
        constraints = [UniqueConstraint(fields=['first', 'second'], name='TaskSchedulingRelationBlueprint_unique_relation')]

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.first == self.second:
            raise ValidationError("First and Second Task Draft must be different.")
        if (str(self.placement) == SchedulingRelationPlacement.Choices.BEFORE.value or str(self.placement) == SchedulingRelationPlacement.Choices.AFTER.value) and self.time_offset<0:
            raise ValidationError("Time_offset must be >= 0")
        super().save(force_insert, force_update, using, update_fields)
    

class TaskSchedulingRelationDraft(BasicCommon, ProjectPropertyMixin):
    first = ForeignKey('TaskDraft', related_name='first_scheduling_relation', on_delete=CASCADE, help_text='First Task Draft to connect.')
    second = ForeignKey('TaskDraft', related_name='second_scheduling_relation', on_delete=CASCADE, help_text='Second Task Draft to connect.')
    placement = ForeignKey('SchedulingRelationPlacement', null=False, on_delete=PROTECT, help_text='Task scheduling relation placement.')
    time_offset = IntegerField(default=60, help_text='Time offset of start of second task with respect to start of first task.')  
    path_to_project = 'first__scheduling_unit_draft__scheduling_set__project'

    class Meta(BasicCommon.Meta):
        # ensure there are no duplicate scheduling relations between tasks
        constraints = [UniqueConstraint(fields=['first', 'second'], name='TaskSchedulingRelationDraft_unique_relation')]

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.first == self.second:
            raise ValidationError("First and Second Task Draft must be different.")
        if (str(self.placement) == SchedulingRelationPlacement.Choices.BEFORE.value or str(self.placement) == SchedulingRelationPlacement.Choices.AFTER.value) and self.time_offset<0:
            raise ValidationError("Time_offset must be >= 0")
        super().save(force_insert, force_update, using, update_fields)


class Reservation(NamedCommon, TemplateSchemaMixin):
    project = ForeignKey('Project', null=True, related_name='reservations', on_delete=CASCADE, help_text='Reservation will be accounted for this project.')
    description = CharField(max_length=255, help_text='Short description for this reservation, used in overviews')
    start_time = DateTimeField(help_text='Start of this reservation.')
    stop_time = DateTimeField(null=True, help_text='Stop of this reservation. If null, then this reservation is indefinitely.')
    specifications_doc = JSONField(help_text='Properties of this reservation')
    specifications_template = ForeignKey('ReservationTemplate', on_delete=CASCADE, help_text='Schema used for specifications_doc.')

    @property
    def duration(self) -> int:
        '''return the overall duration (in seconds) of this task, if stop_time in None than duration ia also None
        '''
        if self.stop_time:
            return (self.stop_time - self.start_time).total_seconds()
        else:
            return None

    def _prevent_reserving_stations_that_are_used_in_active_units(self):
        # Determine active scheduling units that overlap in time with the reservation
        # Note: we cannot filter for status and on sky times in SQL because these are properties
        # todo: find a way to reduce the initial queryset reasonably
        subs = SchedulingUnitBlueprint.objects.all()
        active_subs = [x for x in subs if (x.status in [SchedulingUnitBlueprint.Status.OBSERVING.value, SchedulingUnitBlueprint.Status.SCHEDULED.value]
                                       and x.scheduled_on_sky_stop_time >= self.start_time)]
        if self.stop_time:
            active_subs = [x for x in active_subs if x.scheduled_on_sky_start_time <= self.stop_time]

        # Raise an exception if any of these scheduling units uses a station that shall be reserved
        if "resources" in self.specifications_doc and "stations" in self.specifications_doc["resources"]:
            stations_to_reserve = self.specifications_doc['resources']['stations']
            for sub in active_subs:
                # todo: not all the specified stations may actually be in use. Consider only those who were assigned in the end?
                conflicting_stations = list(set(sub.flat_station_list).intersection(stations_to_reserve))
                if len(conflicting_stations) != 0:
                    msg = "Station(s) %s cannot be reserved because they are currently in use by scheduling unit %s" % (conflicting_stations, sub)
                    logger.info(msg)
                    raise ValueError(msg)

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self._prevent_reserving_stations_that_are_used_in_active_units()
        self.annotate_validate_add_defaults_to_doc_using_template('specifications_doc', 'specifications_template')
        super().save(force_insert, force_update, using, update_fields)


class SystemEvent(NamedCommon):
    created_by = ForeignKey(User, null=False, on_delete=PROTECT, help_text='User who created this entry.')
    start = DateTimeField(help_text='When this event started.')
    stop = DateTimeField(null=True, help_text='When this event stopped (NULLable).')
    issue_type = ForeignKey('SystemEventType', on_delete=PROTECT, help_text='The main type that classifies this issue')
    issue_subtype = ForeignKey('SystemEventSubtype', on_delete=PROTECT, help_text='The subtype that classifies this issue')
    severity = ForeignKey('SystemEventSeverity', on_delete=PROTECT, help_text='The subtype that classifies this issue')
    notes = CharField(max_length=255, null=True, help_text='Any additional information')
    status = ForeignKey('SystemEventStatus', on_delete=PROTECT, help_text='The current status of this issue')
    affected_hardware_doc = JSONField(help_text='Properties of this event, e.g. station list and other hardware affected')
    affected_hardware_template = ForeignKey('SystemEventTemplate', on_delete=PROTECT, help_text='Template for the affected_hardware_doc.')
    affected_tasks = ManyToManyField('TaskBlueprint', blank=True, related_name='system_events', help_text='The task blueprints that are affected by this issue')
    jira_url = CharField(max_length=255, null=True, help_text='Link to JIRA issue (if any)')