"""
This file contains customized UI elements for use in the viewsets (based on the elsewhere defined data models and serializers)
"""
from rest_framework import serializers, fields
import requests
import re
import json

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.common import json_utils
from lofar.common.json_utils import get_default_json_object_for_schema

class JSONEditorField(serializers.JSONField):
    """
    An enhanced JSONField that provides a nice editor widget with validation against the $schema in the json field value.
    """
    def __init__(self, schema_source: str=None, *args, **kwargs):
        '''schema_source should be a string 'pointing to'  the used template and it's schema property.
        For example in the SubtaskSerializer, we point to the specifications_template's schema like so:
          specifications_doc = JSONEditorField(schema_source='specifications_template.schema')
        '''
        self.schema_source = schema_source
        super().__init__(*args, **kwargs)

    def get_schema(self, json_data=None):
        '''get the schema that this JSONEditorField is using via the schema_source'''
        try:
            if isinstance(self.parent.instance, list):
                # the serializer is serializing a list of model instances, so we cannot get a single schema from a single instance
                return None
            return fields.get_attribute(self.parent.instance, self.schema_source.split('.'))
        except (AttributeError, TypeError):
            try:
                if json_data and '$schema' in json_data:
                    schema_url = json_data['$schema']
                    response = requests.get(schema_url)
                    if response.status_code == 200:
                        schema = response.text
                        return json.loads(schema)
            except (KeyError, TypeError, json.JSONDecodeError):
                pass
        return None

    def to_representation(self, value):
        '''create representation of the json-schema-value,
        with all common json schema $ref's pointing to the correct host,
        and inject the josdejong_jsoneditor_widget.html in the render style based on the requests accepted_media_type for single instances'''
        self.style = {}

        if self.parent.context['request'].accepted_media_type == 'text/html' and \
                not (self.parent.parent is not None and hasattr(self.parent.parent, 'many') and self.parent.parent.many):
            # get the used schema...
            schema = self.get_schema(value)

            if schema:
                # ...and 'massage' the served schema such that our rendered html json-editor understands it.
                # the josdejong_jsoneditor_widget cannot resolve absolute URL's in the schema
                # although this should be possible according to the JSON schema standard.
                # so, let's do the resolving here and feed the resolved schema to the josdejong_jsoneditor_widget
                schema = json_utils.resolved_remote_refs(schema)

                # the editor already fetched and cached common meta schema's from json-schema.org
                # and raises an error if we supply it as well
                # so, set schema to None for those
                if 'json-schema.org' in value.get('$schema', ''):
                    schema = None

                self.style = {'template': 'josdejong_jsoneditor_widget.html',
                              'schema': json.dumps(schema)}

        return super().to_representation(value)
