from lofar.sas.tmss.tmss.exceptions import *
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.subtasks import unschedule_subtasks_in_task_blueprint, cancel_subtask
from lofar.sas.tmss.tmss.tmssapp.models.specification import TaskBlueprint, SchedulingUnitBlueprint, SchedulingUnitDraft, IOType, TaskTemplate, TaskType, TaskRelationSelectionTemplate
from lofar.sas.tmss.tmss.tmssapp.subtasks import create_subtasks_from_task_blueprint, schedule_independent_subtasks_in_task_blueprint, update_subtasks_start_times_for_scheduling_unit
from lofar.common.datetimeutils import round_to_minute_precision
from functools import cmp_to_key
import os
from copy import deepcopy
import logging
from datetime import datetime, timedelta
from django.db.utils import IntegrityError
from django.db import transaction
from django.db.models import Q
from lofar.common.util import dict_with_overrides, subdict_of_pointer_items

logger = logging.getLogger(__name__)


def create_scheduling_unit_draft_from_observing_strategy_template(strategy_template: models.SchedulingUnitObservingStrategyTemplate, scheduling_set: models.SchedulingSet, name: str, description: str=None, specifications_doc_overrides: dict=None) -> models.SchedulingUnitDraft:
    '''create a new SchedulingUnitDraft from the given strategy_template with 'parent' scheduling_set'''
    specifications_doc = strategy_template.template_doc_complete_with_defaults

    # apply overrides on template specifications_doc if given
    if specifications_doc_overrides:
        # only use allowed overrides
        allowed_parameter_pointers = sum([p['refs'] for p in strategy_template.template['parameters']], [])
        allowed_overrides = subdict_of_pointer_items(specifications_doc_overrides, allowed_parameter_pointers)

        if allowed_overrides != specifications_doc_overrides:
            raise InvalidSpecificationException("Specified faulty overrides. These are the allowed overridable parameters: \n%s\nSubmitted overrides:%s" % (
                                                '\n'.join([str(p) for p in allowed_parameter_pointers]), specifications_doc_overrides))

        # do the actual override of the remaining allowed items
        specifications_doc = dict_with_overrides(specifications_doc, allowed_overrides)

    if 'scheduling_constraints_template' in specifications_doc:
        scheduling_constraints_template_name_version = specifications_doc.pop('scheduling_constraints_template')
        scheduling_constraints_template = models.SchedulingConstraintsTemplate.objects.get(name=scheduling_constraints_template_name_version['name'],
                                                                                           version=scheduling_constraints_template_name_version.get('version',1))
    else:
        # use the latest scheduling_constraints_template (can be None, which is acceptable)
        scheduling_constraints_template = models.SchedulingConstraintsTemplate.objects.all().order_by('created_at').last()

    # extract the scheduling_constraints_doc from the template_doc,
    # so we can feed in seperately into the SchedulingUnitDraft.
    scheduling_constraints_doc = specifications_doc.pop('scheduling_constraints_doc', {})

    with transaction.atomic():
        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(name=name,
                                                                          description=description,
                                                                          scheduling_set=scheduling_set,
                                                                          specifications_template=strategy_template.scheduling_unit_template,
                                                                          observation_strategy_template=strategy_template,
                                                                          scheduling_constraints_doc=scheduling_constraints_doc,
                                                                          scheduling_constraints_template=scheduling_constraints_template)

        # instantiate the tasks graph
        scheduling_unit_draft = update_task_graph_from_specifications_doc(scheduling_unit_draft, specifications_doc)

        return scheduling_unit_draft


def create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    """
    Create a SchedulingUnitBlueprint from the SchedulingUnitDraft
    :raises Exception if instantiate fails.
    """
    logger.debug("create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft.id=%s name='%s')", scheduling_unit_draft.pk, scheduling_unit_draft.name)

    scheduling_unit_blueprint = SchedulingUnitBlueprint.objects.create(
        name=scheduling_unit_draft.name,
        description=scheduling_unit_draft.description,
        draft=scheduling_unit_draft,
        specifications_template=scheduling_unit_draft.specifications_template,
        ingest_permission_required=scheduling_unit_draft.ingest_permission_required,
        piggyback_allowed_tbb=scheduling_unit_draft.piggyback_allowed_tbb,
        piggyback_allowed_aartfaac=scheduling_unit_draft.piggyback_allowed_aartfaac,
        interrupts_telescope=scheduling_unit_draft.interrupts_telescope)

    logger.info("create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft.id=%s name='%s') created scheduling_unit_blueprint id=%s name='%s'",
                scheduling_unit_draft.pk, scheduling_unit_draft.name, scheduling_unit_blueprint.pk, scheduling_unit_blueprint.name)
    return scheduling_unit_blueprint


def copy_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft, scheduling_set_dst: models.SchedulingSet) -> models.SchedulingUnitDraft:
    """
    Copy a SchedulingUnitDraft
    :raises Exception if instantiate fails.
    """
    with transaction.atomic():
        scheduling_unit_draft_copy = models.SchedulingUnitDraft.objects.create(name="%s (Copy)" % (scheduling_unit_draft.name,),
                                                                               description="%s (Copy from draft '%s' id=%s)" % (scheduling_unit_draft.description or "<no description>", scheduling_unit_draft.name, scheduling_unit_draft.id),
                                                                               scheduling_set=scheduling_set_dst,
                                                                               specifications_template=scheduling_unit_draft.specifications_template,
                                                                               observation_strategy_template=scheduling_unit_draft.observation_strategy_template,
                                                                               scheduling_constraints_doc=scheduling_unit_draft.scheduling_constraints_doc,
                                                                               scheduling_constraints_template=scheduling_unit_draft.scheduling_constraints_template,
                                                                               ingest_permission_required=scheduling_unit_draft.ingest_permission_required,
                                                                               piggyback_allowed_tbb=scheduling_unit_draft.piggyback_allowed_tbb,
                                                                               piggyback_allowed_aartfaac=scheduling_unit_draft.piggyback_allowed_aartfaac,
                                                                               priority_rank=scheduling_unit_draft.priority_rank,
                                                                               priority_queue=scheduling_unit_draft.priority_queue,
                                                                               interrupts_telescope=False # we explicitly set this flag to False while copying, because only real submitted triggers are allowed to set this flag.
                                                                               )

        logger.info("copy_scheduling_unit_draft(scheduling_unit_draft.id=%s) created copy_scheduling_unit_draft id=%s",
                    scheduling_unit_draft.pk, scheduling_unit_draft_copy.pk)

        # instantiate a copy of the task graph
        scheduling_unit_draft_copy = update_task_graph_from_specifications_doc(scheduling_unit_draft_copy, scheduling_unit_draft.specifications_doc)
        return scheduling_unit_draft_copy


def create_scheduling_unit_draft_from_scheduling_unit_blueprint(scheduling_unit_blueprint: models.SchedulingUnitBlueprint) -> models.SchedulingUnitDraft:
    """
    Create a SchedulingUnitDraft from the SchedulingUnitBlueprint
     :raises Exception if instantiate fails.
    """
    logger.debug("create_scheduling_unit_draft_from_scheduling_unit_blueprint(scheduling_unit_blueprint.id=%s)", scheduling_unit_blueprint.pk)

    with transaction.atomic():
        scheduling_unit_draft_copy = models.SchedulingUnitDraft.objects.create(name="%s (Copy from blueprint)" % (scheduling_unit_blueprint.name,),
                                                                               description="%s (Copy from blueprint '%s' id=%s)" % (scheduling_unit_blueprint.description or "<no description>", scheduling_unit_blueprint.name,scheduling_unit_blueprint.id),
                                                                               scheduling_set=scheduling_unit_blueprint.draft.scheduling_set,
                                                                               observation_strategy_template=scheduling_unit_blueprint.draft.observation_strategy_template,
                                                                               specifications_template=scheduling_unit_blueprint.specifications_template,
                                                                               scheduling_constraints_doc=scheduling_unit_blueprint.scheduling_constraints_doc,
                                                                               scheduling_constraints_template=scheduling_unit_blueprint.scheduling_constraints_template,
                                                                               ingest_permission_required=scheduling_unit_blueprint.ingest_permission_required,
                                                                               piggyback_allowed_tbb=scheduling_unit_blueprint.piggyback_allowed_tbb,
                                                                               piggyback_allowed_aartfaac=scheduling_unit_blueprint.piggyback_allowed_aartfaac,
                                                                               priority_rank=scheduling_unit_blueprint.priority_rank,
                                                                               priority_queue=scheduling_unit_blueprint.priority_queue,
                                                                               interrupts_telescope=False # we explicitly set this flag to False while copying, because only real submitted triggers are allowed to set this flag.
                                                                               )

        logger.info("create_scheduling_unit_draft_from_scheduling_unit_blueprint(scheduling_unit_blueprint.id=%s) created copy_scheduling_unit_draft id=%s",
                    scheduling_unit_blueprint.pk, scheduling_unit_draft_copy.pk)

        # instantiate a copy of the tasks graph
        scheduling_unit_draft_copy = update_task_graph_from_specifications_doc(scheduling_unit_draft_copy, scheduling_unit_blueprint.specifications_doc)
        return scheduling_unit_draft_copy


def copy_task_draft(task_draft: models.TaskDraft) -> models.TaskDraft:
    '''create a copy of the given task. Also copy all the task_relation(s) and/or the task_scheduling_relation(s)'''
    logger.debug("copy_task_draft(task_draft.id=%s) copying...", task_draft.pk)

    with transaction.atomic():
        task_draft_copy = models.TaskDraft.objects.create(name="%s (Copy)" % (task_draft.name,),
                                                          description="%s (Copy from task_draft id=%s)" % (task_draft.description, task_draft.id),
                                                          short_description=task_draft.short_description,
                                                          specifications_doc=task_draft.specifications_doc,
                                                          scheduling_unit_draft=task_draft.scheduling_unit_draft,
                                                          specifications_template=task_draft.specifications_template)

        for task_rel in models.TaskRelationDraft.objects.filter(Q(producer__id=task_draft.id) | Q(consumer__id=task_draft.id)).all():
            models.TaskRelationDraft.objects.create(producer=task_draft_copy if task_rel.producer.id==task_draft.id else task_rel.producer,
                                                    consumer=task_draft_copy if task_rel.consumer.id==task_draft.id else task_rel.consumer,
                                                    input_role=task_rel.input_role,
                                                    output_role=task_rel.output_role,
                                                    selection_doc=task_rel.selection_doc,
                                                    selection_template=task_rel.selection_template)

        for task_sched_rel in models.TaskSchedulingRelationDraft.objects.filter(Q(first__id=task_draft.id) | Q(second_id=task_draft.id)).all():
            models.TaskSchedulingRelationDraft.objects.create(first=task_draft_copy if task_sched_rel.first.id==task_draft.id else task_sched_rel.first,
                                                              second=task_draft_copy if task_sched_rel.second.id==task_draft.id else task_sched_rel.second,
                                                              placement=task_sched_rel.placement,
                                                              time_offset=task_sched_rel.time_offset)

        logger.info("copy_task_draft(task_draft.id=%s) created copy_task_draft id=%s", task_draft.pk, task_draft_copy.pk)
        return task_draft_copy


def copy_task_blueprint_to_task_draft(task_blueprint: models.TaskBlueprint) -> models.TaskDraft:
    '''create a (draft)copy of the given task_blueprint. Also copy the task_relation(s) and/or the task_scheduling_relation(s)'''
    logger.debug("copy_task_blueprint_to_task_draft(task_blueprint.id=%s) copying...", task_blueprint.pk)

    with transaction.atomic():
        task_draft_copy = models.TaskDraft.objects.create(name="%s (Copy)" % (task_blueprint.name,),
                                                          description="%s (Copy from task_blueprint id=%s)" % (task_blueprint.description, task_blueprint.id),
                                                          short_description=task_blueprint.short_description,
                                                          specifications_doc=task_blueprint.specifications_doc,
                                                          scheduling_unit_draft=task_blueprint.scheduling_unit_blueprint.draft,
                                                          specifications_template=task_blueprint.specifications_template)

        for task_rel in models.TaskRelationDraft.objects.filter(Q(producer__id=task_blueprint.id) | Q(consumer__id=task_blueprint.id)).all():
            models.TaskRelationDraft.objects.create(producer=task_draft_copy if task_rel.producer.id==task_blueprint.id else task_rel.producer,
                                                    consumer=task_draft_copy if task_rel.consumer.id==task_blueprint.id else task_rel.consumer,
                                                    input_role=task_rel.input_role,
                                                    output_role=task_rel.output_role,
                                                    selection_doc=task_rel.selection_doc,
                                                    selection_template=task_rel.selection_template)

        for task_sched_rel in models.TaskSchedulingRelationDraft.objects.filter(Q(first__id=task_blueprint.id) | Q(second_id=task_blueprint.id)).all():
            models.TaskSchedulingRelationDraft.objects.create(first=task_draft_copy if task_sched_rel.first.id==task_blueprint.id else task_sched_rel.first,
                                                              second=task_draft_copy if task_sched_rel.second.id==task_blueprint.id else task_sched_rel.second,
                                                              placement=task_sched_rel.placement,
                                                              time_offset=task_sched_rel.time_offset)

        logger.info("copy_task_blueprint_to_task_draft(task_blueprint.id=%s) created copy_task_draft id=%s", task_blueprint.pk, task_draft_copy.pk)
        return task_draft_copy


def copy_task_blueprint_via_task_draft_to_new_task_blueprint(task_blueprint: models.TaskBlueprint) -> models.TaskBlueprint:
    '''create a new (blueprint)copy of the given task_blueprint via a new (draft)copy. Also copies all the task_relation(s) and/or the task_scheduling_relation(s)'''
    logger.debug("copy_task_blueprint_via_task_draft_to_new_task_blueprint(task_blueprint.id=%s)", task_blueprint.pk)

    with transaction.atomic():
        task_draft_copy = copy_task_blueprint_to_task_draft(task_blueprint)
        task_blueprint_copy = create_task_blueprint_from_task_draft(task_draft_copy)
        return task_blueprint_copy


def update_task_graph_from_specifications_doc(scheduling_unit_draft: models.SchedulingUnitDraft, specifications_doc: dict) -> models.SchedulingUnitDraft:
    """
    Update the current task graph in this scheduling_unit_draft according to the given specifications_doc: remove tasks from the graph that are not in the doc, update tasks with the same name, create tasks that are in the doc but not in the graph yet.
    """
    logger.debug("update_task_graph_from_specifications_doc(scheduling_unit_draft.id=%s, name='%s') ...", scheduling_unit_draft.pk, scheduling_unit_draft.name)

    with transaction.atomic():
        # remove existing task relations which are not specified (anymore).
        for task_rel in models.TaskRelationDraft.objects.filter(Q(producer__scheduling_unit_draft__id=scheduling_unit_draft.id) |
                                                                Q(consumer__scheduling_unit_draft__id=scheduling_unit_draft.id)).all():
            if len([spec_task_rel for spec_task_rel in specifications_doc.get("task_relations", []) if spec_task_rel['producer']==task_rel.producer.name and spec_task_rel['consumer']==task_rel.consumer.name])==0:
                # existing task_rel is not specified (anymore). Remove it.
                task_rel.delete()
                logger.info("deleted obsolete task_relation_draft id=%s producer=%s,'%s' consumer=%s,'%s' from scheduling_unit_draft id=%s name='%s'",
                            task_rel.id, task_rel.producer.id, task_rel.producer.name, task_rel.consumer.id, task_rel.consumer.name, scheduling_unit_draft.id, scheduling_unit_draft.name)

        # remove existing task scheduling relations which are not specified (anymore).
        for task_sched_rel in models.TaskSchedulingRelationDraft.objects.filter(Q(first__scheduling_unit_draft__id=scheduling_unit_draft.id) |
                                                                                Q(first__scheduling_unit_draft__id=scheduling_unit_draft.id)).all():
            if len([spec_task_sched_rel for spec_task_sched_rel in specifications_doc.get("task_scheduling_relations",[]) if spec_task_sched_rel['first']==task_sched_rel.first.name and spec_task_sched_rel['second']==task_sched_rel.second.name])==0:
                # existing task_sched_rel is not specified (anymore). Remove it.
                task_sched_rel.delete()
                logger.info("deleted obsolete task_scheduling_relation_draft id=%s first=%s,'%s' second=%s,'%s' from scheduling_unit_draft id=%s name='%s'",
                            task_sched_rel.id, task_sched_rel.first.id, task_sched_rel.first.name, task_sched_rel.second.id, task_sched_rel.second.name, scheduling_unit_draft.id, scheduling_unit_draft.name)

        # remove existing tasks which are not specified (anymore).
        for task in scheduling_unit_draft.task_drafts.all():
            if len([spec_task_name for spec_task_name in specifications_doc.get("tasks",{}).keys() if spec_task_name==task.name])==0:
                # existing task_sched_rel is not specified (anymore). Remove it.
                task.delete()
                logger.info("deleted obsolete task_draft id=%s name='%s' from scheduling_unit_draft id=%s name='%s'",
                            task.id, task.name, scheduling_unit_draft.id, scheduling_unit_draft.name)

        # add/update new/updated tasks
        for task_name, task_definition in specifications_doc.get("tasks", {}).items():
            task_template_name = task_definition["specifications_template"]["name"]
            task_template_version = task_definition["specifications_template"].get("version", 1)
            task_template = models.TaskTemplate.objects.get(name=task_template_name, version=task_template_version)

            task_specifications_doc = task_definition["specifications_doc"]
            task_specifications_doc = task_template.add_defaults_to_json_object_for_schema(task_specifications_doc)

            logger.debug("creating/updating task draft... task_name='%s', task_template_name='%s', task_template_version=%s", task_name, task_template_name, task_template_version)

            try:
                task_draft = scheduling_unit_draft.task_drafts.get(name=task_name)
                task_draft.description = task_definition.get("description", "")
                task_draft.short_description = task_definition.get("short description", "")
                task_draft.specifications_doc = task_specifications_doc
                task_draft.specifications_template = task_template
                task_draft.save()

                logger.info("updated task draft id=%s task_name='%s', task_template_name='%s', task_template_version=%s",
                            task_draft.pk, task_name, task_template_name, task_template_version)
            except models.TaskDraft.DoesNotExist:
                task_draft = models.TaskDraft.objects.create(name=task_name,
                                                             description=task_definition.get("description", ""),
                                                             short_description=task_definition.get("short description", ""),
                                                             scheduling_unit_draft=scheduling_unit_draft,
                                                             specifications_doc = task_specifications_doc,
                                                             specifications_template=task_template)
                logger.info("created task draft id=%s task_name='%s', task_template_name='%s', task_template_version=%s",
                            task_draft.pk, task_name, task_template_name, task_template_version)

        scheduling_unit_draft.refresh_from_db()

        # Now create task relations
        for task_relation_definition in specifications_doc.get("task_relations", []):
            try:
                producer_task_draft = scheduling_unit_draft.task_drafts.get(name=task_relation_definition["producer"])
                consumer_task_draft = scheduling_unit_draft.task_drafts.get(name=task_relation_definition["consumer"])
                input_role = models.TaskConnectorType.objects.get(task_template=consumer_task_draft.specifications_template,
                                                                  role=task_relation_definition["input"]["role"],
                                                                  datatype=task_relation_definition["input"]["datatype"],
                                                                  dataformat=task_relation_definition["input"]["dataformat"],
                                                                  iotype=models.IOType.Choices.INPUT.value)
                output_role = models.TaskConnectorType.objects.get(task_template=producer_task_draft.specifications_template,
                                                                   role=task_relation_definition["output"]["role"],
                                                                   datatype=task_relation_definition["output"]["datatype"],
                                                                   dataformat=task_relation_definition["output"]["dataformat"],
                                                                   iotype=models.IOType.Choices.OUTPUT.value)
                selection_template = models.TaskRelationSelectionTemplate.objects.get(name=task_relation_definition["selection_template"])
                selection_doc = task_relation_definition.get("selection_doc", selection_template.get_default_json_document_for_schema())
            except Exception as e:
                logger.error("Could not determine Task Relations for %s. Error: %s", task_relation_definition, e)
                raise

            task_relation, created = models.TaskRelationDraft.objects.update_or_create(tags=task_relation_definition.get("tags",[]),
                                                                                       producer=producer_task_draft,
                                                                                       consumer=consumer_task_draft,
                                                                                       input_role=input_role,
                                                                                       output_role=output_role,
                                                                                       selection_doc=selection_doc,
                                                                                       selection_template=selection_template)
            logger.info("%s task_relation id=%s between task draft id=%s name='%s' and id=%s name='%s",
                        "created" if created else "updated",
                        task_relation.pk, producer_task_draft.id, producer_task_draft.name, consumer_task_draft.id, consumer_task_draft.name)

        # task_scheduling_relation
        for task_scheduling_relation_definition in specifications_doc.get("task_scheduling_relations", []):
            placement = models.SchedulingRelationPlacement.objects.get(value=task_scheduling_relation_definition["placement"])
            time_offset = task_scheduling_relation_definition["time_offset"]
            first_task_draft = scheduling_unit_draft.task_drafts.get(name=task_scheduling_relation_definition["first"])
            second_task_draft = scheduling_unit_draft.task_drafts.get(name=task_scheduling_relation_definition["second"])

            task_scheduling_relation, created = models.TaskSchedulingRelationDraft.objects.update_or_create(placement=placement,
                                                                                                            time_offset=time_offset,
                                                                                                            first=first_task_draft,
                                                                                                            second=second_task_draft)
            logger.info("%s task_scheduling_relation id=%s between task draft id=%s name='%s' and id=%s name='%s",
                        "created" if created else "updated",
                        task_scheduling_relation.pk, first_task_draft.id, first_task_draft.name, second_task_draft.id, second_task_draft.name)

    scheduling_unit_draft.refresh_from_db()
    return scheduling_unit_draft


def create_task_blueprint_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
    """
    Create a task_blueprint from the task_draft
    :raises Exception if instantiate fails.
    """
    logger.debug("creating task_blueprint from task_draft id=%s", task_draft.pk)

    with transaction.atomic():
        # get or create a scheduling_unit_blueprint from the scheduling_unit_draft
        scheduling_unit_blueprint = task_draft.scheduling_unit_draft.scheduling_unit_blueprints.last()
        if scheduling_unit_blueprint is None:
            scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(task_draft.scheduling_unit_draft)

        try:
            task_blueprint = TaskBlueprint.objects.get(name=task_draft.name, scheduling_unit_blueprint=scheduling_unit_blueprint)
            if task_blueprint.draft != task_draft:
                task_blueprint.draft = task_draft
                task_blueprint.save()
        except TaskBlueprint.DoesNotExist:
            task_blueprint = TaskBlueprint.objects.create(description=task_draft.description,
                                                          short_description=task_draft.short_description,
                                                          name=task_draft.name,
                                                          draft=task_draft,
                                                          scheduling_unit_blueprint=scheduling_unit_blueprint,
                                                          specifications_doc=task_draft.specifications_doc,
                                                          specifications_template=task_draft.specifications_template,
                                                          output_pinned=task_draft.output_pinned)

            logger.info("created task_blueprint id=%s from task_draft id=%s name=%s", task_blueprint.pk, task_draft.pk, task_draft.name)

        # now that we have a task_blueprint, its time to refresh the task_draft so we get the non-cached fields
        task_draft.refresh_from_db()

        # loop over consumers/producers, and 'copy' the TaskRelationBlueprint from the TaskRelationDraft
        # this is only possible if both 'ends' of the task_relation are converted to a TaskBlueprint
        # so, when converting two TaskDrafts (for example an observation and a pipeline), then for the conversion
        # of the first TaskDraft->TaskBlueprint no relation is setup,
        # and for the second TaskDraft->TaskBlueprint conversion we have both endpoints available, so we can connect them.
        task_draft_relations = models.TaskRelationDraft.objects.filter(Q(consumer=task_draft) | Q(producer=task_draft)).all()
        for task_relation_draft in task_draft_relations:
            try:
                producing_task_blueprint = task_relation_draft.producer.task_blueprints.get(scheduling_unit_blueprint=task_blueprint.scheduling_unit_blueprint)
                consuming_task_blueprint = task_relation_draft.consumer.task_blueprints.get(scheduling_unit_blueprint=task_blueprint.scheduling_unit_blueprint)

                with transaction.atomic():
                    task_relation_blueprint = models.TaskRelationBlueprint.objects.create(draft=task_relation_draft,
                                                                                          input_role=task_relation_draft.input_role,
                                                                                          output_role=task_relation_draft.output_role,
                                                                                          producer=producing_task_blueprint,
                                                                                          consumer=consuming_task_blueprint,
                                                                                          selection_doc=task_relation_draft.selection_doc,
                                                                                          selection_template=task_relation_draft.selection_template)
                    logger.info("created task_relation_blueprint id=%s which connects task_blueprints producer_id=%s and consumer_id=%s",
                                task_relation_blueprint.pk, producing_task_blueprint.pk, consuming_task_blueprint.pk)
            except models.TaskBlueprint.DoesNotExist:
                # either the producer or the consumer has no task_blueprint (yet)
                pass
            except IntegrityError as e:
                if 'TaskRelationBlueprint_unique_relation' in str(e):
                    logger.info("task_relation_blueprint with producer_id=%s and consumer_id=%s already exists",
                                producing_task_blueprint.pk, consuming_task_blueprint.pk)
                else:
                    raise


        # Do the same 'trick' for Task Scheduling Relation Draft to Blueprint
        task_draft_scheduling_relations = models.TaskSchedulingRelationDraft.objects.filter(Q(first=task_draft) | Q(second=task_draft)).all()
        for task_scheduling_relation_draft in task_draft_scheduling_relations:
            try:
                first_task_blueprint = task_scheduling_relation_draft.first.task_blueprints.get(scheduling_unit_blueprint=task_blueprint.scheduling_unit_blueprint)
                second_task_blueprint = task_scheduling_relation_draft.second.task_blueprints.get(scheduling_unit_blueprint=task_blueprint.scheduling_unit_blueprint)

                with transaction.atomic():
                    task_scheduling_relation_blueprint = models.TaskSchedulingRelationBlueprint.objects.create(draft=task_scheduling_relation_draft,
                                                                                                               first=first_task_blueprint,
                                                                                                               second=second_task_blueprint,
                                                                                                               time_offset=task_scheduling_relation_draft.time_offset,
                                                                                                               placement=task_scheduling_relation_draft.placement)
                    logger.info("created task_scheduling_relation_blueprint id=%s which connects task_blueprints first_id=%s and second_id=%s, placement=%s time_offset=%s[sec]",
                                task_scheduling_relation_blueprint.pk, first_task_blueprint.pk, second_task_blueprint.pk, task_scheduling_relation_draft.placement, task_scheduling_relation_draft.time_offset)
            except models.TaskBlueprint.DoesNotExist:
                # either the first or the second has no task_blueprint (yet)
                pass
            except IntegrityError as e:
                if 'TaskSchedulingRelationBlueprint_unique_relation' in str(e):
                    logger.info("task_scheduling_relation_blueprint with first_id=%s and second_id=%s already exists",
                                first_task_blueprint.pk, second_task_blueprint.pk)
                else:
                    raise

        return task_blueprint


def create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Create the scheduling_unit_blueprint, then create its child task_blueprint(s), then create the task_blueprint's subtasks'''
    with transaction.atomic():
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft)
        return update_task_blueprints_and_subtasks_graph_from_draft(scheduling_unit_blueprint)


def create_task_blueprint_and_subtasks_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
    '''Convenience method: Create the task_blueprint, then create the task_blueprint's subtasks, and schedule the ones that are not dependend on predecessors'''
    with transaction.atomic():
        task_blueprint =  create_task_blueprint_from_task_draft(task_draft)
        create_subtasks_from_task_blueprint(task_blueprint)
    task_blueprint.refresh_from_db()
    return task_blueprint


def update_task_blueprint_graph_from_draft(scheduling_unit_blueprint: models.SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    '''Update this scheduling_unit_blueprint's task_blueprints graph from the linked scheduling_unit_draft's graph.
    - New task_drafts that are not in the blueprint graph (yet) are added to this blueprint graph.
    - Modified task_drafts that are in the blueprint graph are NOT updated to the task_blueprint, because task_blueprints are immutable.
    - Deleted task_drafts that are in the blueprint graph are NOT removed from the blueprint graph, because task_blueprints are immutable and carved in stone and reflect what happened.
    '''

    with transaction.atomic():
        task_drafts = list(scheduling_unit_blueprint.draft.task_drafts.all())

        # sort them in 'data-flow'-order,
        # because successors can depend on predecessors, so the first taskdraft's need to be blueprinted first.
        task_drafts.sort(key=cmp_to_key(lambda taskdraft_a, taskdraft_b: -1 if taskdraft_a in taskdraft_b.predecessors else 1 if taskdraft_b in taskdraft_a.predecessors else 0))

        # convert task_draft(s) to task_blueprint(s)
        for task_draft in task_drafts:
            create_task_blueprint_from_task_draft(task_draft)

        # refresh so all related fields are updated.
        scheduling_unit_blueprint.refresh_from_db()
        return scheduling_unit_blueprint


def update_task_blueprints_and_subtasks_graph_from_draft(scheduling_unit_blueprint: models.SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Create the scheduling_unit_blueprint's task_blueprint(s), then create each task_blueprint's subtasks'''
    with transaction.atomic():
        scheduling_unit_blueprint = update_task_blueprint_graph_from_draft(scheduling_unit_blueprint)

        for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
            create_subtasks_from_task_blueprint(task_blueprint)

        # assign reasonable default start/stop times so the subtasks/tasks/sched_unit can be displayed in a timeline view
        # these default start/stop times can of course be overridden by the operator and/or dynamic scheduling.
        update_subtasks_start_times_for_scheduling_unit(scheduling_unit_blueprint, round_to_minute_precision(datetime.utcnow()+timedelta(hours=1)))

    # refresh so all related fields are updated.
    scheduling_unit_blueprint.refresh_from_db()

    return scheduling_unit_blueprint


def schedule_independent_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint: SchedulingUnitBlueprint, start_time: datetime=None) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Schedule the subtasks in the scheduling_unit_blueprint that are not dependend on predecessors'''
    task_blueprints = list(scheduling_unit_blueprint.task_blueprints.all())

    for task_blueprint in task_blueprints:
        schedule_independent_subtasks_in_task_blueprint(task_blueprint, start_time=start_time+task_blueprint.relative_start_time)

    scheduling_unit_blueprint.refresh_from_db()
    return scheduling_unit_blueprint


def unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint: SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Unschedule all scheduled subtasks in the scheduling_unit_blueprint'''
    task_blueprints = list(scheduling_unit_blueprint.task_blueprints.all())

    for task_blueprint in task_blueprints:
        unschedule_subtasks_in_task_blueprint(task_blueprint)

    scheduling_unit_blueprint.refresh_from_db()
    return scheduling_unit_blueprint


def cancel_task_blueprint(task_blueprint: TaskBlueprint) -> TaskBlueprint:
    '''Convenience method: cancel all subtasks in the task_blueprint'''
    for subtask in task_blueprint.subtasks.all():
        cancel_subtask(subtask)
    task_blueprint.refresh_from_db()
    return task_blueprint


def mark_task_blueprint_as_obsolete(task_blueprint: TaskBlueprint) -> TaskBlueprint:
    '''Convenience method: mark all cancelled/error subtasks in the task_blueprint as obsolete'''
    now = datetime.utcnow()
    for subtask in task_blueprint.subtasks.all():
        subtask.obsolete_since = now
        subtask.save()
    task_blueprint.refresh_from_db()
    return task_blueprint


def cancel_scheduling_unit_blueprint(scheduling_unit_blueprint: SchedulingUnitBlueprint) -> SchedulingUnitBlueprint:
    '''Convenience method: cancel all subtasks in the task_blueprints in the scheduling_unit_blueprint'''
    for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
        cancel_task_blueprint(task_blueprint)
    scheduling_unit_blueprint.refresh_from_db()
    return scheduling_unit_blueprint


def mark_scheduling_unit_blueprint_as_obsolete(scheduling_unit_blueprint: SchedulingUnitBlueprint) -> SchedulingUnitBlueprint:
    '''Convenience method: mark all subtasks in the task_blueprints in the scheduling_unit_blueprint as obsolete'''
    for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
        mark_task_blueprint_as_obsolete(task_blueprint)
    scheduling_unit_blueprint.refresh_from_db()
    return scheduling_unit_blueprint



def create_cleanuptask_for_scheduling_unit_blueprint(scheduling_unit_blueprint: SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    '''create a cleanuptask for the given scheduling_unit which will cleanup all output dataproducts from tasks in this scheduling_unit which aren't already cleaned up'''

    # Rationale:
    # adding a cleanup task(blueprint) to a scheduling_unit_blueprint adds a task to the graph (which breaks the immutable blueprint concept),
    # but it does not modify observation/pipeline behaviour, hence we allow it.
    # Regard this as a convenience function to allow users to simplify cleaning up after themselves if they forgot to specificy a cleanup task.

    with transaction.atomic():
        # create a cleanup task draft and blueprint....
        cleanup_template = models.TaskTemplate.objects.get(name="cleanup")
        cleanup_spec_doc = cleanup_template.get_default_json_document_for_schema()

        cleanup_task_blueprint = TaskBlueprint.objects.create(
            name="Cleanup",
            description="Cleaning up all output dataproducts for this scheduling unit",
            scheduling_unit_blueprint=scheduling_unit_blueprint,
            specifications_doc=cleanup_spec_doc,
            specifications_template=cleanup_template,
            output_pinned=False)

        logger.info("Created Cleanup Task id=%d for scheduling_unit id=%s, adding the outputs of all producing tasks in the scheduling unit to the cleanup...", cleanup_task_blueprint.id, scheduling_unit_blueprint.id)

        # ... and connect the outputs of the producing tasks to the cleanup, so the cleanup task knows what to remove.
        selection_template = TaskRelationSelectionTemplate.objects.get(name="all")
        selection_doc = selection_template.get_default_json_document_for_schema()

        for producer_task_blueprint in scheduling_unit_blueprint.task_blueprints.exclude(specifications_template__type=TaskType.Choices.CLEANUP).exclude(specifications_template__type=TaskType.Choices.INGEST).all():
            for connector_type in producer_task_blueprint.specifications_template.connector_types.filter(iotype__value=IOType.Choices.OUTPUT.value).all():
                # define what the producer_task_blueprint is producing
                output_role = models.TaskConnectorType.objects.get(task_template=producer_task_blueprint.specifications_template,
                                                                   role=connector_type.role,
                                                                   datatype=connector_type.datatype,
                                                                   iotype=models.IOType.objects.get(value=models.IOType.Choices.OUTPUT.value))

                # define what the cleanup task accepts/consumes
                input_role = models.TaskConnectorType.objects.filter(dataformat=connector_type.dataformat).get(task_template=cleanup_task_blueprint.specifications_template,
                                                                                                               role=models.Role.objects.get(value=models.Role.Choices.ANY.value),
                                                                                                               datatype=connector_type.datatype,
                                                                                                               iotype=models.IOType.objects.get(value=models.IOType.Choices.INPUT.value))

                # connect the two
                task_relation_blueprint = models.TaskRelationBlueprint.objects.create(producer=producer_task_blueprint,
                                                                                      consumer=cleanup_task_blueprint,
                                                                                      input_role=input_role,
                                                                                      output_role=output_role,
                                                                                      selection_doc=selection_doc,
                                                                                      selection_template=selection_template)

                logger.info("created task_relation id=%s between task blueprint id=%s name='%s' and id=%s name='%s",
                            task_relation_blueprint.pk, task_relation_blueprint.producer.id, task_relation_blueprint.producer.name, task_relation_blueprint.consumer.id, task_relation_blueprint.consumer.name)


        # add the same to the draft graph
        copy_task_blueprint_to_task_draft(cleanup_task_blueprint)

        # and finally also create the executable subtask for the cleanup_task_blueprint, so it can actually run.
        create_subtasks_from_task_blueprint(cleanup_task_blueprint)

        # return the modified scheduling_unit
        scheduling_unit_blueprint.refresh_from_db()
        return scheduling_unit_blueprint


def create_cleanuptask_for_scheduling_unit_draft(scheduling_unit_draft: SchedulingUnitDraft) -> models.SchedulingUnitDraft:
    '''create a cleanuptask for the given scheduling_unit which will cleanup all output dataproducts from tasks in this scheduling_unit which aren't already cleaned up'''

    with transaction.atomic():
        # create a cleanup task draft....
        cleanup_template = models.TaskTemplate.objects.get(name="cleanup")
        cleanup_spec_doc = cleanup_template.get_default_json_document_for_schema()

        cleanup_task_draft = models.TaskDraft.objects.create(
            name="Cleanup",
            description="Cleaning up all output dataproducts for this scheduling unit",
            scheduling_unit_draft=scheduling_unit_draft,
            specifications_doc=cleanup_spec_doc,
            specifications_template=cleanup_template)

        logger.info("Created Cleanup Task id=%d for scheduling_unit id=%s, adding the outputs of all producing tasks in the scheduling unit to the cleanup...", cleanup_task_draft.id, scheduling_unit_draft.id)

        # ... and connect the outputs of the producing tasks to the cleanup, so the cleanup task knows what to remove.
        selection_template = TaskRelationSelectionTemplate.objects.get(name="all")
        selection_doc = selection_template.get_default_json_document_for_schema()

        for producer_task_draft in scheduling_unit_draft.task_drafts.exclude(specifications_template__type=TaskType.Choices.CLEANUP).exclude(specifications_template__type=TaskType.Choices.INGEST).all():
            for connector_type in producer_task_draft.specifications_template.connector_types.filter(iotype__value=IOType.Choices.OUTPUT.value).all():
                # define what the producer_task_draft is producing
                output_role = models.TaskConnectorType.objects.get(task_template=producer_task_draft.specifications_template,
                                                                   role=connector_type.role,
                                                                   datatype=connector_type.datatype,
                                                                   iotype=models.IOType.objects.get(value=models.IOType.Choices.OUTPUT.value))

                # define what the cleanup task accepts/consumes
                input_role = models.TaskConnectorType.objects.filter(dataformat=connector_type.dataformat).get(task_template=cleanup_task_draft.specifications_template,
                                                                                                               role=models.Role.objects.get(value=models.Role.Choices.ANY.value),
                                                                                                               datatype=connector_type.datatype,
                                                                                                               iotype=models.IOType.objects.get(value=models.IOType.Choices.INPUT.value))

                # connect the two
                task_relation_draft = models.TaskRelationDraft.objects.create(producer=producer_task_draft,
                                                                              consumer=cleanup_task_draft,
                                                                              input_role=input_role,
                                                                              output_role=output_role,
                                                                              selection_doc=selection_doc,
                                                                              selection_template=selection_template)

                logger.info("created task_relation id=%s between task draft id=%s name='%s' and id=%s name='%s",
                            task_relation_draft.pk, task_relation_draft.producer.id, task_relation_draft.producer.name, task_relation_draft.consumer.id, task_relation_draft.consumer.name)

        # return the modified scheduling_unit
        scheduling_unit_draft.refresh_from_db()
        return scheduling_unit_draft
