from collections import OrderedDict
from django.utils.encoding import force_str
from rest_framework.metadata import SimpleMetadata

import logging
logger = logging.getLogger(__name__)


class LOFARDefaultMetadata(SimpleMetadata):
    """
    Metadata class to include filtering and ordering info for `OPTIONS` requests.
    """
    def determine_metadata(self, request, view):
        metadata = super(LOFARDefaultMetadata, self).determine_metadata(request, view)

        metadata['filters'] = OrderedDict()

        if hasattr(view, 'filter_fields'):
            all_fields = view.get_serializer().Meta.model._meta.fields
            filters = OrderedDict()
            if view.filter_fields == '__all__':
                # Iterate over all fields
                for f in all_fields:
                    filters[f.name] = {"type": str(f.__class__.__name__), "lookup_types": ["exact"]}
                metadata['filters'] = {**metadata['filters'], **filters}
            else:
                for f in view.filter_fields:
                    for af in all_fields:
                        if f == af.name:
                            filters[af.name] = {"type": str(af.__class__.__name__), "lookup_types": ["exact"]}
                metadata['filters'] = {**metadata['filters'], **filters}

        if hasattr(view, 'filter_class'):
            # Add filters
            filters = OrderedDict()
            for filter_name, filter_type in view.filter_class.base_filters.items():
                filter_parts = filter_name.split('__')
                filter_name = filter_parts[0]
                attrs = OrderedDict()

                # Type
                attrs['type'] = filter_type.__class__.__name__

                # Lookup fields
                if len(filter_parts) > 1:
                    # Has a lookup type (__gt, __lt, etc.)
                    lookup_type = filter_parts[1]
                    if filters.get(filter_name) is not None:
                        # Filter name already in, just append the value
                        attrs['lookup_types'] = filters[filter_name]['lookup_types']
                        attrs['lookup_types'].append(lookup_type)
                    else:
                        attrs['lookup_types'] = [lookup_type]
                else:
                    attrs['lookup_types'] = ['exact']

                # Do choices
                choices = filter_type.extra.get('choices', False)
                if choices:
                    attrs['choices'] = [
                        {
                            'value': choice_value,
                            'display_name': force_str(choice_name, strings_only=True)
                        }
                        for choice_value, choice_name in choices
                    ]
                filters[filter_name] = attrs

            # Add filters info
            metadata['filters'] = {**metadata['filters'], **filters}

        # Add ordering info
        # NOTE: If you don't specify an ordering attribute on the view, the filter class will default to allowing
        # the user to filter on any readable fields on the serializer specified by the serializer_class attribute.
        if hasattr(view, 'ordering_fields'):
            if view.ordering_fields == '__all__':
                # Iterate over all fields
                fields = view.get_serializer().Meta.model._meta.fields
                metadata['ordering'] = [f.name for f in fields]
            else:
                metadata['ordering'] = [f for f in view.ordering_fields]

        return metadata
