"""
This file contains the viewsets (based on the elsewhere defined data models and serializers)
"""

from django.shortcuts import get_object_or_404
from rest_framework import viewsets
from .lofar_viewset import LOFARViewSet, LOFARNestedViewSet, AbstractTemplateViewSet, FILTER_OVERRIDES
from .specification import NumberInFilter
from .. import models
from .. import serializers
from django_filters import rest_framework as filters
from lofar.sas.tmss.tmss.tmssapp.models.scheduling import Subtask
# Don't use Ordering class from the django_filters but use rest_framework instead !!
from rest_framework.filters import OrderingFilter
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from drf_yasg.inspectors import SwaggerAutoSchema
from drf_yasg.openapi import Parameter

from rest_framework.decorators import action
from rest_framework.decorators import permission_classes
from django.http import HttpResponse, JsonResponse, HttpResponseRedirect, HttpResponseNotFound
from rest_framework.response import Response as RestResponse

from lofar.common import isProductionEnvironment, isTestEnvironment
from lofar.sas.tmss.tmss.tmssapp.viewsets.lofar_viewset import LOFARViewSet
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp import serializers
from lofar.sas.tmss.tmss.exceptions import TMSSException

from datetime import datetime
from lofar.common.json_utils import get_default_json_object_for_schema
from lofar.common.datetimeutils import formatDatetime
from lofar.sas.tmss.tmss.tmssapp.adapters.parset import convert_to_parset
from drf_yasg.renderers import _SpecRenderer
import json

from lofar.sas.tmss.tmss.tmssapp.renderers import PlainTextRenderer
from rest_framework.views import APIView
from rest_framework.decorators import api_view, renderer_classes
from django.core.exceptions import ObjectDoesNotExist
import django_property_filter as property_filters
from django.db import transaction

class TextPlainAutoSchema(SwaggerAutoSchema):
    def get_produces(self):
        return ["text/plain"]


class SubtaskStateViewSet(LOFARViewSet):
    queryset = models.SubtaskState.objects.all()
    serializer_class = serializers.SubtaskStateSerializer


class SubtaskAllowedStateTransitionsViewSet(LOFARViewSet):
    queryset = models.SubtaskAllowedStateTransitions.objects.all()
    serializer_class = serializers.SubtaskAllowedStateTransitionsSerializer


class SubtaskStateLogViewSet(LOFARViewSet):
    queryset = models.SubtaskStateLog.objects.all()
    serializer_class = serializers.SubtaskStateLogSerializer

    def get_queryset(self):
        if 'subtask_id' in self.kwargs:
            subtask = get_object_or_404(models.Subtask, pk=self.kwargs['subtask_id'])
            return subtask.subtaskstatelog_set.order_by('created_at').all()

        queryset = models.SubtaskStateLog.objects.all()

        # query by subtask
        subtask = self.request.query_params.get('subtask', None)
        if subtask is not None:
            return queryset.filter(subtask=subtask)

        return queryset


class SubtaskTypeViewSet(LOFARViewSet):
    queryset = models.SubtaskType.objects.all()
    serializer_class = serializers.SubtaskTypeSerializer

class StationTypeViewSet(LOFARViewSet):
    queryset = models.StationType.objects.all()
    serializer_class = serializers.StationTypeSerializer

class HashAlgorithmViewSet(LOFARViewSet):
    queryset = models.HashAlgorithm.objects.all()
    serializer_class = serializers.HashAlgorithmSerializer

class SubtaskTemplateFilter(filters.FilterSet):
    class Meta:
        model = models.SubtaskTemplate
        fields = {
            'name': ['exact'],
            'version': ['lt', 'gt', 'exact']
        }

class SubtaskTemplateViewSet(AbstractTemplateViewSet):
    queryset = models.SubtaskTemplate.objects.all()
    serializer_class = serializers.SubtaskTemplateSerializer
    filter_class = SubtaskTemplateFilter

    def get_queryset(self):
        queryset = models.SubtaskTemplate.objects.all()

        # query by name
        name = self.request.query_params.get('name', None)
        if name is not None:
            return queryset.filter(name=name)

        return queryset


class DataproductSpecificationsTemplateViewSet(AbstractTemplateViewSet):
    queryset = models.DataproductSpecificationsTemplate.objects.all()
    serializer_class = serializers.DataproductSpecificationsTemplateSerializer


class DataproductFeedbackTemplateViewSet(AbstractTemplateViewSet):
    queryset = models.DataproductFeedbackTemplate.objects.all()
    serializer_class = serializers.DataproductFeedbackTemplateSerializer


class SubTaskFilter(property_filters.PropertyFilterSet):
    id = NumberInFilter(field_name='id', lookup_expr='in')
    id_min = filters.NumberFilter(field_name='id', lookup_expr='gte')
    id_max = filters.NumberFilter(field_name='id', lookup_expr='lte')
    state = filters.ModelMultipleChoiceFilter(field_name='state', queryset=models.SubtaskState.objects.all())
    name = filters.CharFilter(field_name='task_blueprint__scheduling_unit_blueprint__name', lookup_expr='icontains')   # todo: correct name?
    error_reason = filters.CharFilter(field_name='error_reason', lookup_expr='icontains')
    on_sky_start_time__lt = property_filters.PropertyIsoDateTimeFilter(field_name='on_sky_start_time', lookup_expr='lt')
    on_sky_start_time__gt = property_filters.PropertyIsoDateTimeFilter(field_name='on_sky_start_time', lookup_expr='gt')
    on_sky_stop_time__lt = property_filters.PropertyIsoDateTimeFilter(field_name='on_sky_stop_time', lookup_expr='lt')
    on_sky_stop_time__gt = property_filters.PropertyIsoDateTimeFilter(field_name='on_sky_stop_time', lookup_expr='gt')
    process_start_time__lt = property_filters.PropertyIsoDateTimeFilter(field_name='process_start_time', lookup_expr='lt')
    process_start_time__gt = property_filters.PropertyIsoDateTimeFilter(field_name='process_start_time', lookup_expr='gt')
    process_stop_time__lt = property_filters.PropertyIsoDateTimeFilter(field_name='process_stop_time', lookup_expr='lt')
    process_stop_time__gt = property_filters.PropertyIsoDateTimeFilter(field_name='process_stop_time', lookup_expr='gt')
    scheduled_on_sky_start_time__lt = filters.IsoDateTimeFilter(field_name='scheduled_on_sky_start_time', lookup_expr='lt')
    scheduled_on_sky_start_time__gt = filters.IsoDateTimeFilter(field_name='scheduled_on_sky_start_time', lookup_expr='gt')
    scheduled_on_sky_stop_time__lt = filters.IsoDateTimeFilter(field_name='scheduled_on_sky_stop_time', lookup_expr='lt')
    scheduled_on_sky_stop_time__gt = filters.IsoDateTimeFilter(field_name='scheduled_on_sky_stop_time', lookup_expr='gt')
    actual_on_sky_start_time__lt = filters.IsoDateTimeFilter(field_name='actual_on_sky_start_time', lookup_expr='lt')
    actual_on_sky_start_time__gt = filters.IsoDateTimeFilter(field_name='actual_on_sky_start_time', lookup_expr='gt')
    actual_on_sky_stop_time__lt = filters.IsoDateTimeFilter(field_name='actual_on_sky_stop_time', lookup_expr='lt')
    actual_on_sky_stop_time__gt = filters.IsoDateTimeFilter(field_name='actual_on_sky_stop_time', lookup_expr='gt')
    scheduled_process_start_time__lt = filters.IsoDateTimeFilter(field_name='scheduled_process_start_time', lookup_expr='lt')
    scheduled_process_start_time__gt = filters.IsoDateTimeFilter(field_name='scheduled_process_start_time', lookup_expr='gt')
    scheduled_process_stop_time__lt = filters.IsoDateTimeFilter(field_name='scheduled_process_stop_time', lookup_expr='lt')
    scheduled_process_stop_time__gt = filters.IsoDateTimeFilter(field_name='scheduled_process_stop_time', lookup_expr='gt')
    actual_process_start_time__lt = filters.IsoDateTimeFilter(field_name='actual_process_start_time', lookup_expr='lt')
    actual_process_start_time__gt = filters.IsoDateTimeFilter(field_name='actual_process_start_time', lookup_expr='gt')
    actual_process_stop_time__lt = filters.IsoDateTimeFilter(field_name='actual_process_stop_time', lookup_expr='lt')
    actual_process_stop_time__gt = filters.IsoDateTimeFilter(field_name='actual_process_stop_time', lookup_expr='gt')

    subtask_type = filters.ModelMultipleChoiceFilter(field_name='specifications_template__type', queryset=models.SubtaskType.objects.all())

    class Meta:
        model = Subtask
        fields = {
            'state__value': ['exact'],
            'cluster__name': ['exact', 'icontains'],
        }
        filter_overrides = FILTER_OVERRIDES


class SubtaskViewSet(LOFARViewSet):
    queryset = models.Subtask.objects.all()
    serializer_class = serializers.SubtaskSerializer
    filter_class = SubTaskFilter
    ordering = ('scheduled_on_sky_start_time',)

    # performance boost: select the related models in a single db call.
    queryset = queryset.select_related('state', 'specifications_template', 'specifications_template__type', 'cluster', 'created_or_updated_by_user')

    @swagger_auto_schema(auto_schema=TextPlainAutoSchema,
                         responses={200: 'A LOFAR parset for this subtask (as plain text)',
                                    403: 'forbidden',
                                    404: 'Not found'},
                         produces='text/plain',
                         operation_description="Get a LOFAR parset for the specifications of this subtask")
    @action(methods=['get'], detail=True, renderer_classes=[PlainTextRenderer])
    def parset(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        parset = convert_to_parset(subtask)

        header = "# THIS PARSET WAS GENERATED BY TMSS FROM THE SPECIFICATION OF SUBTASK ID=%d ON %s\n" % (subtask.pk, formatDatetime(datetime.utcnow()))
        parset_str = header + str(parset).replace('"','').replace("'","") # remove quotes
        return HttpResponse(parset_str, content_type='text/plain')


    @swagger_auto_schema(responses={200: 'The predecessor subtasks of this subtask',
                                    403: 'forbidden'},
                         operation_description="Get the predecessor subtasks of this subtask.")
    @action(methods=['get'], detail=True, url_name="predecessors")
    def predecessors(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        predecessors = self.filter_queryset(subtask.predecessors)
        serializer = self.get_serializer(predecessors, many=True)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The successor subtasks of this subtask',
                                    403: 'forbidden'},
                         operation_description="Get the successor subtasks of this subtask.")
    @action(methods=['get'], detail=True, url_name="successors")
    def successors(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        successors = self.filter_queryset(subtask.successors)
        serializer = self.get_serializer(successors, many=True)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The scheduled version of this subtask',
                                    403: 'forbidden',
                                    500: 'The subtask could not be scheduled'},
                         operation_description="Try to schedule this subtask.")
    @action(methods=['post'], detail=True, url_name="schedule")
    def schedule(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.subtasks import schedule_subtask_and_update_successor_start_times
        scheduled_subtask = schedule_subtask_and_update_successor_start_times(subtask)
        serializer = self.get_serializer(scheduled_subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The unscheduled version of this subtask',
                                    403: 'forbidden',
                                    500: 'The subtask could not be unscheduled'},
                         operation_description="Try to unschedule this subtask, deleting all the output dataproducts and setting status back to 'defined'.")
    @action(methods=['post'], detail=True, url_name="unschedule")
    def unschedule(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.subtasks import unschedule_subtask
        unscheduled_subtask = unschedule_subtask(subtask)
        serializer = self.get_serializer(unscheduled_subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The cancelled version of this subtask',
                                    403: 'forbidden',
                                    500: 'The subtask could not be cancelled'},
                         operation_description="Try to cancel this subtask.")
    @action(methods=['post'], detail=True, url_name="cancel")
    def cancel(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.subtasks import cancel_subtask
        cancelled_subtask = cancel_subtask(subtask)
        serializer = self.get_serializer(cancelled_subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The state log for this Subtask.',
                                    403: 'forbidden'},
                         operation_description="Get the state log for this Subtask.")
    @action(methods=['get'], detail=True)
    def state_log(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        serializer = serializers.SubtaskStateLogSerializer(subtask.subtaskstatelog_set.order_by('created_at').all(),
                                                           many=True,
                                                           context={'request': request})
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={302: 'A redirect url to the task log for this Subtask.',
                                    403: 'forbidden'},
                         operation_description="Get the task log for this Subtask.")
    @action(methods=['get'], detail=True)
    def task_log(self, request, pk=None):
        """
        Return a redirect to the the link to the pipeline log in case of pipeline or
        link to COBALT error log in case of an observation.
        """
        subtask = get_object_or_404(models.Subtask, pk=pk)

        # redirect to cobalt log served at proxy.lofar.eu
        if subtask.specifications_template.type.value == models.SubtaskType.Choices.OBSERVATION.value:
            url = "https://proxy.lofar.eu/inspect/%s/rtcp-%s.errors" % (subtask.id, subtask.id)
            return HttpResponseRedirect(redirect_to=url)

        # redirect to pipeline log served via webscheduler
        if subtask.specifications_template.type.value == models.SubtaskType.Choices.PIPELINE.value:
            # import here and not at top of module to "loosen" dependency on external packages, such as in this case the RADB RPC.
            from lofar.sas.resourceassignment.resourceassignmentservice.rpc import RADBRPC
            # Get RADBID, subtask must be at least 'scheduled' to exist in radb
            try:
                with RADBRPC.create(timeout=2) as radbrpc:
                    radb_id = radbrpc.getTask(tmss_id=subtask.id)

                    if radb_id is None:
                        return HttpResponseNotFound(
                            content='No RADB task found for subtask id=%s type="%s status=%s". Cannot redirect to pipeline log.' % (
                                subtask.id, subtask.specifications_template.type.value, subtask.state))

                    WEBSCHEDULER_URL = "http://scu001.control.lofar:7412" if isProductionEnvironment() else \
                                       "http://scu199.control.lofar:7412" if isTestEnvironment() else \
                                       "http://localhost:7412"

                    url = "%s/tasks/%s/log.html" % (WEBSCHEDULER_URL, radb_id)
                    return HttpResponseRedirect(redirect_to=url)
            except Exception as e:
                return HttpResponseNotFound(content='No RADB task found for subtask id=%s type="%s". Cannot redirect to pipeline log.' % (subtask.id, subtask.specifications_template.type.value))

        # unknown log
        return HttpResponseNotFound(content='No log (url) available for subtask id=%s type="%s"' % (subtask.id, subtask.specifications_template.type.value) )


    @swagger_auto_schema(responses={200: 'The input dataproducts of this subtask.',
                                    403: 'forbidden'},
                         operation_description="Get the input dataproducts of this subtask.")
    @action(methods=['get'], detail=True, url_name='input_dataproducts')
    def input_dataproducts(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        # fetch the subtask's input_dataproducts and all related properties in one queryset. Major speedup.
        input_dataproducts = subtask.input_dataproducts.select_related('dataformat', 'datatype', 'specifications_template', 'producer', 'feedback_template', 'sap', 'global_identifier', 'archive_info').prefetch_related('consumers')
        serializer = serializers.DataproductSerializer(input_dataproducts, many=True, context={'request': request})
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The output dataproducts of this subtask.',
                                    403: 'forbidden'},
                         operation_description="Get the output dataproducts of this subtask.")
    @action(methods=['get'], detail=True, url_name='output_dataproducts')
    def output_dataproducts(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        # fetch the subtask's output_dataproducts and all related properties in one queryset. Major speedup.
        output_dataproducts = subtask.output_dataproducts.select_related('dataformat', 'datatype', 'specifications_template', 'producer', 'feedback_template', 'sap', 'global_identifier', 'archive_info').prefetch_related('consumers')
        serializer = serializers.DataproductSerializer(output_dataproducts, many=True, context={'request': request})
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The transformed output dataproduct of this subtask for the given input_dataproduct_id.',
                                    403: 'forbidden'},
                         operation_description='Getthe transformed output dataproducts for the given input_dataproduct_id.',
                         manual_parameters=[Parameter(name='input_dataproduct_id', required=True, type='integer', in_='query',
                                                      description="the id of the input dataproduct for which you want to get the transformed output dataproduct")])
    @action(methods=['get'], detail=True, url_name='transformed_output_dataproduct')
    def transformed_output_dataproduct(self, request, pk=None):
        '''return the transformed output dataproduct for the given input_dataproduct_id.'''
        subtask = get_object_or_404(models.Subtask, pk=pk)
        input_dataproduct_id = request.query_params['input_dataproduct_id']
        try:
            output_dataproduct = subtask.get_transformed_output_dataproduct(input_dataproduct_id)
        except models.Dataproduct.DoesNotExist:
            return HttpResponseNotFound('Cannot find transformed output dataproduct for subtask id=%s and input_dataproduct_id=%s.' % (pk, input_dataproduct_id))

        serializer = serializers.DataproductSerializer(output_dataproduct, many=False, context={'request': request})
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The updated (maybe) finished version of this subtask with the posted feedback appended to the raw_feedback property.',
                                    403: 'forbidden',
                                    500: 'The feedback for this subtask could not be processed'},
                         operation_description="Process the feedback_doc (which can be for one or more or all dataproducts), store/append it in the subtask's raw_feedback, and process it into json feedback per dataproduct. Sets the subtask to finished if all dataproducts are processed, which may require multiple postings of partial feedback docs.")
    @action(methods=['post'], detail=True, url_path='process_feedback_and_set_to_finished_if_complete')
    def process_feedback_and_set_to_finished_if_complete(self, request, pk=None):
        from lofar.sas.tmss.tmss.tmssapp.adapters.feedback import process_feedback_for_subtask_and_set_to_finished_if_complete
        subtask = get_object_or_404(models.Subtask, pk=pk)
        feedback_doc = request.body.decode('utf-8')
        updated_subtask = process_feedback_for_subtask_and_set_to_finished_if_complete(subtask, feedback_doc)
        serializer = self.get_serializer(updated_subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The updated (maybe) finished version of this subtask after reprocessing the stored raw_feedback.',
                                    403: 'forbidden',
                                    500: 'The feedback for this subtask could not be reprocessed'},
                         operation_description="Reprocess the raw_feedback in the subtask into json feedback per dataproduct. Sets the subtask to finished if all dataproducts are processed.")
    @action(methods=['get'], detail=True, url_path='reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete')
    def reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete(self, request, pk=None):
        from lofar.sas.tmss.tmss.tmssapp.adapters.feedback import reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete
        subtask = get_object_or_404(models.Subtask, pk=pk)
        updated_subtask = reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete(subtask)
        serializer = self.get_serializer(updated_subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'Get progress of this subtask ranging from 0.0 when just started up to 1.0 when finished.',
                                    403: 'forbidden',
                                    500: 'Could not compute the progress'},
                         operation_description="Get progress of this subtask ranging from 0.0 when just started up to 1.0 when finished.")
    @action(methods=['get'], detail=True, url_name='get_progress')
    def get_progress(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        return JsonResponse({'id': subtask.id, 'progress': subtask.progress})


class SubtaskExtendedViewSet(SubtaskViewSet):
    serializer_class = serializers.SubtaskExtendedSerializer


class SubtaskNestedViewSet(LOFARNestedViewSet):
    queryset = models.Subtask.objects.all()
    serializer_class = serializers.SubtaskSerializer
    filter_class = SubTaskFilter
    ordering = ('scheduled_on_sky_start_time',)

    # performance boost: select the related models in a single db call.
    queryset = queryset.select_related('state', 'specifications_template', 'specifications_template__type', 'cluster', 'created_or_updated_by_user')

    def get_queryset(self):
        if 'task_blueprint_id' in self.kwargs:
            task_blueprint = get_object_or_404(models.TaskBlueprint, pk=self.kwargs['task_blueprint_id'])
            return task_blueprint.subtasks.all()


class SubtaskInputViewSet(LOFARViewSet):
    queryset = models.SubtaskInput.objects.all()
    serializer_class = serializers.SubtaskInputSerializer


class SubtaskInputExtendedViewSet(SubtaskInputViewSet):
    serializer_class = serializers.SubtaskInputExtendedSerializer


class SubtaskOutputViewSet(LOFARViewSet):
    queryset = models.SubtaskOutput.objects.all()
    serializer_class = serializers.SubtaskOutputSerializer


class SubtaskOutputExtendedViewSet(SubtaskOutputViewSet):
    serializer_class = serializers.SubtaskOutputExtendedSerializer


class DataproductViewSet(LOFARViewSet):
    queryset = models.Dataproduct.objects.all()
    serializer_class = serializers.DataproductSerializer

    # performance boost: select the related models in a single db call.
    queryset = queryset.select_related('dataformat', 'datatype', 'specifications_template', 'feedback_template', 'sap', 'global_identifier')

    @swagger_auto_schema(responses={200: 'The SIP for this dataproduct',
                                    403: 'forbidden'},
                         operation_description="Get the Submission Information Package (SIP) for this dataproduct")
    @action(methods=['get'], detail=True, url_name="sip")
    def sip(self, request, pk=None):
        from lofar.sas.tmss.tmss.tmssapp.adapters.sip import generate_sip_for_dataproduct
        from lofar.sas.tmss.tmss.tmssapp import views
        from django.urls import reverse

        # get the dataproduct...
        dataproduct = get_object_or_404(models.Dataproduct, pk=pk)

        # generate the sip
        sip = generate_sip_for_dataproduct(dataproduct).get_prettyxml()

        # construct the schema location for the sip
        lta_sip_xsd_path = reverse(views.get_lta_sip_xsd)
        lta_sip_xsd_uri = "%s://%s/%s" % (request.scheme, request.get_host().rstrip('/'), lta_sip_xsd_path.lstrip('/'))
        # the schema_location is a space seperated string: "<namespace> <schema_url>"
        # the "weird" thing is that according to the XSD standard, the <namespace> should be a URI usually in the form of a URL which does not necessarily need to resolve...
        # We'll accept the weirdness, and use the http://www.astron.nl/SIP-Lofar namespace even though we don't host the XSD there, and we'll inject the true XSD url here so it can actually be fetched by validators
        schema_location = "%s %s" % ("http://www.astron.nl/SIP-Lofar", lta_sip_xsd_uri)

        # ugly string manipulation to inject the schemaLocation attribute (saves expensive XML parsing & DOM manipulation)
        sip = sip.replace('<sip:ltaSip ', '<sip:ltaSip xsi:schemaLocation="%s" ' % schema_location)

        # and return it
        return HttpResponse(sip, content_type='application/xml')

    @swagger_auto_schema(responses={200: 'The SIP graph for this dataproduct',
                                    403: 'forbidden'},
                         operation_description="Get a visualization of the Submission Information Package (SIP) for this dataproduct")
    @action(methods=['get'], detail=True, url_name="sip_graph")
    def sip_graph(self, request, pk=None):
        dataproduct = get_object_or_404(models.Dataproduct, pk=pk)
        from lofar.lta.sip import visualizer
        from lofar.sas.tmss.tmss.tmssapp.adapters.sip import generate_sip_for_dataproduct
        return HttpResponse(visualizer.visualize_sip(generate_sip_for_dataproduct(dataproduct)), content_type='image/svg+xml')

    @swagger_auto_schema(responses={200: 'The updated dataproduct.',
                                    403: 'forbidden',
                                    500: 'The feedback of this subtask could not be processed'},
                         operation_description="Store the archive information for this dataproduct.")
    @action(methods=['post'], detail=True, url_name='post_archive_information')
    def post_archive_information(self, request, pk=None):
        dataproduct = get_object_or_404(models.Dataproduct, pk=pk)
        if dataproduct.producer.subtask.specifications_template.type.value != models.SubtaskType.Choices.INGEST.value:
            raise TMSSException("Cannot store archive information for dataproduct id=%s because its producing subtask id=%s is of non-ingest type=%s" % (pk, dataproduct.producer.subtask.id,
                                                                                                                                                         dataproduct.producer.subtask.specifications_template.type.value))

        json_doc = json.loads(request.body.decode('utf-8'))

        with transaction.atomic():
            dataproduct.size = int(json_doc['file_size'])
            dataproduct.directory, dataproduct.filename = json_doc['srm_url'].rsplit('/', maxsplit=1)

            if 'storage_ticket' in json_doc:
                models.DataproductArchiveInfo.objects.create(dataproduct=dataproduct, storage_ticket=json_doc['storage_ticket'])

            if 'md5_checksum' in json_doc:
                models.DataproductHash.objects.create(dataproduct=dataproduct,
                                                      hash_algorithm=models.HashAlgorithm.objects.get(value=models.HashAlgorithm.Choices.MD5.value),
                                                      hash=json_doc['md5_checksum'])

            if 'adler32_checksum' in json_doc:
                models.DataproductHash.objects.create(dataproduct=dataproduct,
                                                      hash_algorithm=models.HashAlgorithm.objects.get(value=models.HashAlgorithm.Choices.ADLER32.value),
                                                      hash=json_doc['adler32_checksum'])

            # create empty feedback. Apart from the archive info above, ingest does not create feedback like observations/pipelines do.
            dataproduct.feedback_template = models.DataproductFeedbackTemplate.objects.get(name="empty")
            dataproduct.feedback_doc = dataproduct.feedback_template.get_default_json_document_for_schema()
            dataproduct.save()

        serializer = self.get_serializer(dataproduct)
        return RestResponse(serializer.data)


class AntennaSetViewSet(LOFARViewSet):
    queryset = models.AntennaSet.objects.all()
    serializer_class = serializers.AntennaSetSerializer


class DataproductTransformViewSet(LOFARViewSet):
    queryset = models.DataproductTransform.objects.all()
    serializer_class = serializers.DataproductTransformSerializer


class FilesystemViewSet(LOFARViewSet):
    queryset = models.Filesystem.objects.all()
    serializer_class = serializers.FilesystemSerializer


class ClusterViewSet(LOFARViewSet):
    queryset = models.Cluster.objects.all()
    serializer_class = serializers.ClusterSerializer


class DataproductArchiveInfoViewSet(LOFARViewSet):
    queryset = models.DataproductArchiveInfo.objects.all()
    serializer_class = serializers.DataproductArchiveInfoSerializer


class DataproductHashViewSet(LOFARViewSet):
    queryset = models.DataproductHash.objects.all()
    serializer_class = serializers.DataproductHashSerializer


class SAPViewSet(LOFARViewSet):
    queryset = models.SAP.objects.all()
    serializer_class = serializers.SAPSerializer

    @swagger_auto_schema(responses={200: 'The dataproducts in this SAP',
                                    403: 'forbidden'},
                         operation_description="The dataproducts in this SAP.")
    @action(methods=['get'], detail=True, url_name='dataproducts')
    def dataproducts(self, request, pk=None):
        sap = get_object_or_404(models.SAP, pk=pk)
        serializer = serializers.DataproductSerializer(sap.dataproducts, many=True, context={'request': request})
        return RestResponse(serializer.data)


class SAPTemplateViewSet(AbstractTemplateViewSet):
    queryset = models.SAPTemplate.objects.all()
    serializer_class = serializers.SAPTemplateSerializer


class SIPidentifierViewSet(LOFARViewSet):
    queryset = models.SIPidentifier.objects.all()
    serializer_class = serializers.SIPidentifierSerializer
